from optimus.engines.base.commons.functions import is_integer_cudf, is_float_cudf, is_numeric_cudf, is_string_cudf
from optimus.engines.base.mask import Mask


class CUDFMask(Mask):

    def numeric(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_numeric_cudf)

    def int(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_integer_cudf)

    def float(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_float_cudf)

    def str(self, cols="*"):
        return self.root[cols].cols.apply(cols, is_string_cudf)


from optimus.engines.base.create import BaseCreate
from optimus.helpers.types import DataFrameType, InternalDataFrameType
import cudf
import pandas as pd

from optimus.engines.dask_cudf.dataframe import DaskCUDFDataFrame

class Create(BaseCreate):

    def _dfd_from_dict(self, dict) -> InternalDataFrameType:
        return cudf.DataFrame({ name: cudf.Series(values, dtype=dtype) for (name, dtype, nulls), values in dict.items() })

    def _df_from_dfd(self, dfd, n_partitions=1, *args, **kwargs) -> DataFrameType:
        if isinstance(dfd, (pd.DataFrame,)):
            dfd = cudf.from_pandas(dfd)
        return DaskCUDFDataFrame(dfd, n_partitions, *args, **kwargs)



import re
import dask

from optimus.engines.base.functions import Functions
from optimus.helpers.core import val_to_list
import dask.dataframe as dd

class DaskBaseFunctions():
    def word_tokenize(self, series):
        pass

    def delayed(self, func):
        def wrapper(*args, **kwargs):
            return dask.delayed(func)(*args, **kwargs)

        return wrapper

    def from_delayed(self, delayed):
        return dd.from_delayed(delayed)

    def to_delayed(self, value):
        return value.to_delayed()

    def to_float(self, series, *args):
        return series.map_partitions(self._to_float_partition, meta=float)

    def _to_float(self, series):
        return self.to_float(series)

    def _to_integer(self, series):
        return self.to_integer(series)

    def to_integer(self, series, *args):
        return series.map_partitions(self._to_integer_partition, meta=int)

    def to_string(self, series):
        return series.astype(str)

    def min(self, series):
        return series.min()

    def max(self, series):
        return series.max()

    def count_zeros(self, series):
        return int((self._to_float(series).values == 0).sum())

    def replace_chars(self, series, search, replace_by):
        regex=False
        replace_by = val_to_list(replace_by)
        if len(search) == 1:
            _search = search[0]
        else:
            regex=True
            _search = ('|'.join(map(re.escape, search)))
        
        if len(replace_by) <= 1:
            _r = replace_by[0]
        else:
            regex=True
            _map = {s: r for s, r in zip(search, replace_by)}
            def _r(value):
                return _map.get(value[0], "ERROR")

        series = self.to_string_accessor(series).replace(_search, _r, regex=regex)
        return series



# from dask_ml.preprocessing import DummyEncoder

from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_, is_str
import pandas as pd


class Encoding:
    def __init__(self, root):
        self.root = root

    def n_gram(self, input_col, n=2):
        """
        Converts the input array of strings inside of DF into an array of n-grams.
        :param df: Dataframe to analyze
        :param input_col: Column to analyzer.
        :param n: number of elements per n-gram >=1.
        :return:
        """

        pass

    def one_hot_encoder(self, input_cols, prefix=None, **kargs):
        """
        Maps a column of label indices to a column of binary vectors, with at most a single one-value.
        :param df: Dataframe to be transformed.
        :param input_cols: Columns to be encoded.
        :param output_col: Column where the output is going to be saved.
        :return: Dataframe with encoded columns.
        """
        dfd = self.root.data
        df = self.root
        input_cols = parse_columns(df, input_cols)

        return self.root.new(pd.concat([dfd, pd.get_dummies(dfd[input_cols], prefix=prefix)], axis=1))

    # TODO: Must we use the pipeline version?
    def vector_assembler(self, input_cols, output_col=None):
        """
        Combines a given list of columns into a single vector column.
        :param df: Dataframe to be transformed.
        :param input_cols: Columns to be assembled.
        :param output_col: Column where the output is going to be saved.
        :return: Dataframe with assembled column.
        """

        input_cols = parse_columns(df, input_cols)

        if output_col is None:
            output_col = name_col(input_cols, "vector_assembler")

        # df = pipeline.fit(df).transform(df)

        return df

    def normalizer(df, input_cols, output_col=None, p=2.0):
        """
        Transforms a dataset of Vector rows, normalizing each Vector to have unit norm. It takes parameter p, which
        specifies the p-norm used for normalization. (p=2) by default.
        :param df: Dataframe to be transformed
        :param input_cols: Columns to be normalized.
        :param output_col: Column where the output is going to be saved.
        :param p:  p-norm used for normalization.
        :return: Dataframe with normalized columns.
        """

        # Check if columns argument must be a string or list datat ype:
        if not is_(input_cols, (str, list)):
            RaiseIt.type_error(input_cols, ["str", "list"])

        if is_str(input_cols):
            input_cols = [input_cols]

        if is_(input_cols, (float, int)):
            RaiseIt.type_error(input_cols, ["float", "int"])

        # Try to create a vector
        if len(input_cols) > 1:
            df = df.cols.cast(input_cols, "vector")

        if output_col is None:
            output_col = name_col(input_cols, "normalizer")

        # TODO https://developer.ibm.com/code/2018/04/10/improve-performance-ml-pipelines-wide-dataframes-apache-spark-2-3/

        # df = pipeline.fit(df).transform(df)

        return df



import errno
from optimus.api.code import generate_code
import warnings
import os
from io import UnsupportedOperation
from pprint import pformat

from optimus.engines.base.basedataframe import BaseDataFrame

from optimus.infer import is_function, is_list_value, is_list_empty, is_list_of_str, is_list_of_numeric, is_list_of_tuples, \
    is_numeric, is_str, is_dict
from optimus.helpers.debug import get_var_name
from optimus.helpers.logger import logger


class Test:
    def __init__(self, op=None, df=None, name=None, path=None, final_path=None, options={}, **kwargs):
        """
        Create python code with unit test functions for Optimus.
        :param op: optimus instance
        :param df: Spark Dataframe
        :param name: Name of the Test Class
        :type path: folder where tests will be written individually. run() nneds to be runned to merge all the tests.
        :param options: dictionary with configuration options
            import: Libraries to be added. 
            engine: Engine to use. 
            n_partitions: Number of partitions of created dataframes (if supported)

        """
        self.op = op
        self.df = df
        self.name = name
        self.path = path
        self.final_path = final_path
        self.options = options if options != {} else kwargs

    def run(self):
        """
        Return the tests in text format
        :return:
        """

        final_path = self.final_path
        if self.path is None:
            filename = "test_" + self.name + ".py"
        else:
            filename = final_path + "/" + "test_" + self.name + ".py"

        test_file = open(filename, 'w', encoding='utf-8')
        print("Creating file " + filename)

        # Imports
        _imports = [
            "import numpy as np",
            "nan = np.nan", 
            "from optimus import Optimus",
            "from optimus.helpers.json import json_enconding",
            "from optimus.helpers.functions import deep_sort",
            "import unittest"
        ]

        if self.options.get("imports", None) is not None:
            for i in self.options["imports"]:
                _imports.append(i)
        for i in _imports:
            test_file.write(i + "\n")

        # Engine
        engine = self.options.get("engine", "pandas")

        # Number of partitions
        n_partitions = self.options.get("n_partitions", 1)

        # Creating the Optimus instance
        test_file.write(f"op = Optimus(\"{engine}\")\n")

        # Global Dataframe
        if self.df is not None:
            source_df = generate_code(source="op", target="source_df",
                                      operation="create.dataframe", dict=self.df.export(), n_partitions=n_partitions) + "\n"
            test_file.write(source_df)

        # Class name
        cls = "class Test_" + self.name + "(unittest.TestCase):\n"

        test_file.write(cls)
        test_file.write("\tmaxDiff = None\n")

        for root, dirs, files in os.walk(self.path):
            for file in files:
                if file.endswith(".test"):
                    full_path = os.path.join(root, file)

                    with open(full_path, 'r', encoding='utf-8') as opened_file:
                        try:
                            text = opened_file.read()

                            test_file.write(text)
                            opened_file.close()
                        except UnsupportedOperation:
                            print("file seems to be empty")

        test_file.close()
        print("Done")

    def create(self, obj, method, suffix=None, output="df", additional_method=None, *args, **kwargs):
        """
        This is a helper function that output python tests for Spark DataFrames.
        :param obj: Object to be tested
        :param method: Method to be tested
        :param suffix: The test name will be create using the method param. suffix will add a string in case you want
        to customize the test name.
        :param output: can be a 'df' or a 'json'
        :param additional_method:
        :param args: Arguments to be used in the method
        :param kwargs: Keyword arguments to be used in the functions
        :return:
        """

        buffer = []

        def add_buffer(value):
            buffer.append("\t" + value)

        # Create name
        name = []

        if method is not None:
            name.append(method.replace(".", "_"))

        if additional_method is not None:
            name.append(additional_method)

        if suffix is not None:
            name.append(suffix)

        test_name = "_".join(name)

        func_test_name = "test_" + test_name + "()"

        print("Creating {test} test function...".format(test=func_test_name))
        logger.print(func_test_name)

        if not output == "dict":
            add_buffer("@staticmethod\n")
            func_test_name = "test_" + test_name + "()"
        else:
            func_test_name = "test_" + test_name + "(self)"

        filename = test_name + ".test"

        add_buffer("def " + func_test_name + ":\n")

        source = "source_df"

        n_partitions = self.options.get("n_partitions", 1)

        if obj is None:
            # Use the main df
            df_func = self.df
        elif isinstance(obj, (BaseDataFrame,)):

            source_df = "\t" + generate_code(source="op", target="source_df",
                                             operation="create.dataframe", dict=obj.export(), n_partitions=n_partitions) + "\n"

            df_func = obj
            add_buffer(source_df)
        else:
            source = get_var_name(obj)
            df_func = obj

        # Process simple arguments
        _args = []
        for v in args:
            if is_function(v):
                _args.append(v.__qualname__)
            elif isinstance(v, (BaseDataFrame,)):
                _df = generate_code(source="op", target=False, operation="create.dataframe", dict=v.export(),
                                    n_partitions=n_partitions)
                _args.append(_df)
            else:
                _args.append(pformat(v))

        _args = ','.join(_args)
        _kwargs = []

        # Process keywords arguments
        for k, v in kwargs.items():
            _kwargs.append(k + "=" + pformat(v))

        # Separator if we have positional and keyword arguments
        separator = ""
        if (not is_list_empty(args)) & (not is_list_empty(kwargs)):
            separator = ","

        if method is None:
            add_buffer("\tactual_df = source_df\n")
        else:
            am = ""
            if additional_method:
                am = "." + additional_method + "()"

            add_buffer("\tactual_df =" + source + "." + method + "(" + _args + separator + ','.join(
                _kwargs) + ")" + am + "\n")

        # print("df_result", df_result)
        # Apply function to the spark
        if method is None:
            df_result = self.op.create.dataframe(
                *args, **kwargs, n_partitions=n_partitions)
        else:
            # Here we construct the method to be applied to the source object
            for f in method.split("."):
                df_func = getattr(df_func, f)

            df_result = df_func(*args, **kwargs)

        # Additional Methods
        if additional_method is not None:
            df_result = getattr(df_result, additional_method)()
        if output == "df":
            df_result.table()
            expected = "\t" + \
                generate_code(source="op", target="expected_df", operation="create.dataframe", dict=df_result.export(), n_partitions=n_partitions) + \
                "\n"

        elif output == "json":
            if is_str(df_result):
                df_result = "'" + df_result + "'"
            else:
                df_result = str(df_result)
            add_buffer("\tactual_df = json_enconding(actual_df)\n")
            expected = "\texpected_value = json_enconding(" + df_result + ")\n"

        elif output == "dict":
            expected = "\texpected_value =" + str(df_result) + "\n"
        else:
            expected = "\t\n"

        add_buffer(expected)

        # Output
        if output == "df":
            add_buffer("\tassert (expected_df.collect() == actual_df.collect())\n")
        elif output == "json":
            add_buffer("\tassert(expected_value == actual_df)\n")
        elif output == "dict":
            add_buffer("\tself.assertDictEqual(deep_sort(expected_value),  deep_sort(actual_df))\n")

        filename = self.path + "//" + filename
        if not os.path.exists(os.path.dirname(filename)):
            try:
                os.makedirs(os.path.dirname(filename))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise

        # Write file
        test_file = open(filename, 'w', encoding='utf-8')

        for b in buffer:
            test_file.write(b)

        # return "".join(buffer)

    def delete(self, df, func, suffix=None, *args, **kwargs):
        """
        This is a helper function that delete python tests files used to construct the final Test file.
        :param df: Do nothing, only for simplicity so you can delete a file test the same way you create it
        :param suffix: The create method will try to create a test function with the func param given.
        If you want to test a function with different params you can use suffix.
        :param func: Function to be tested
        :return:
        """

        if suffix is None:
            suffix = ""
        else:
            suffix = "_" + suffix

        # Create func test name. If is None we just test the create.df function a not transform the data frame in
        # any way
        if func is None:
            func_test_name = "test_" + "create_df" + suffix + "()"
            filename = "create_df" + suffix + ".test"

        else:
            func_test_name = "test_" + func.replace(".", "_") + suffix + "()"

            filename = func.replace(".", "_") + suffix + ".test"

        filename = self.path + "//" + filename

        print("Deleting file {test}...".format(test=filename))
        logger.print(func_test_name)
        try:
            os.remove(filename)
        except FileNotFoundError as e:
            warnings.warn(e)



import copy

from glom import glom, assign, delete

from optimus.helpers.core import val_to_list
from optimus.infer import is_list_value

ACTIONS_PATH = "transformations.actions"


class Meta:

    @staticmethod
    def set(meta, spec=None, value=None, missing=dict) -> dict:
        """
        Set metadata in a dataframe columns
        :param meta: Meta data to be modified
        :param spec: path to the key to be modified
        :param value: dict value
        :param missing:
        :return:
        """
        if spec is not None:
            data = copy.deepcopy(meta)
            assign(data, spec, value, missing=missing)
        else:
            data = value

        return data

    @staticmethod
    def reset(meta, spec=None) -> dict:
        """
        Set metadata in a dataframe columns
        :param meta: Meta data to be modified
        :param spec: path to the key to be modified
        :param value: dict value
        :param missing:
        :return:
        """
        if spec is not None:
            data = copy.deepcopy(meta)
            delete(data, spec, ignore_missing=True)
        else:
            data = meta

        return data

    @staticmethod
    def get(meta, spec=None) -> dict:
        """
        Get metadata from a dataframe column
        :param meta:Meta data to be modified
        :param spec: path to the key to be modified
        :return: dict
        """
        if spec is not None:
            data = glom(meta, spec, skip_exc=KeyError)
        else:
            data = meta
        return copy.deepcopy(data)

    @staticmethod
    def reset_actions(meta):
        """
        Reset the data frame metadata
        :param meta: Meta data to be modified
        :return:
        """

        return Meta.set(meta, ACTIONS_PATH, [])


    @staticmethod
    def columns(meta, value) -> dict:
        """
        Shortcut to cache the columns in a dataframe
        :param meta: Meta data to be modified
        :param value:
        :return: dict (Meta)
        """
        value = val_to_list(value)
        for v in value:
            meta = Meta.update(meta, "transformations.columns", v, list)
        return meta

    @staticmethod
    def action(meta, name, value) -> dict:
        """
        Shortcut to add actions to a dataframe
        :param meta: Meta data to be modified
        :param name: Action name
        :param value: Value to be added
        :return: dict (Meta)
        """
        if not is_list_value(value):
            value = [value]
        for v in value:
            meta = Meta.update(meta, ACTIONS_PATH, {name: v}, list)
        return meta

    @staticmethod
    def update(meta, path, value, default=list) -> dict:
        """
        Update meta data in a key
        :param meta: Meta data to be modified
        :param path: Path indise the dict to be modified
        :param value: New key value
        :param default:
        :return: dict (Meta)
        """

        new_meta = copy.deepcopy(meta)

        if new_meta is None:
            new_meta = {}

        elements = path.split(".")
        _element = new_meta
        for i, ele in enumerate(elements):
            if ele not in _element and not len(elements) - 1 == i:
                _element[ele] = {}

            if len(elements) - 1 == i:
                if default is list:
                    _element[ele] = _element.get(ele, [])
                    _element[ele].append(value)
                elif default is dict:
                    _element[ele] = _element.get(ele, {})
                    _element[ele].update(value)
            else:
                _element = _element[ele]

        return new_meta


import re
import inspect
from inspect import signature
from pprint import pformat

from optimus.engines.base.basedataframe import BaseDataFrame as dataframe_class
from optimus.infer import is_list, is_str
from optimus.helpers.types import DataFrameType, ConnectionType, ClustersType, MaskDataFrameType, is_any_optimus_type
from optimus.helpers.core import val_to_list, one_list_to_val

from .prepare import prepare

from optimus.engines.base.engine import BaseEngine as engine_class
from optimus.engines.base.stringclustering import Clusters as clusters_class

from optimus.engines.base.create import BaseCreate as create_class
from optimus.engines.base.io.load import BaseLoad as load_class
from optimus.engines.base.io.save import BaseSave as save_class
from optimus.engines.base.io.connect import Connect as connect_class

from optimus.engines.base.columns import BaseColumns as cols_class
from optimus.engines.base.rows import BaseRows as rows_class
from optimus.engines.base.set import BaseSet as set_class
from optimus.engines.base.mask import Mask as mask_class
from optimus.engines.base.ml.models import ML as ml_class
from optimus.plots.plots import Plot as plots_class
from optimus.outliers.outliers import Outliers as outliers_class
from optimus.engines.base.profile import BaseProfile as profile_class

engine_accessors = {
    "create": create_class,
    "load": load_class,
    "save": save_class,
    "connect": connect_class
}

dataframe_accessors = {
    "cols": cols_class,
    "rows": rows_class,
    "set": set_class,
    "mask": mask_class,
    "ml": ml_class,
    "plot": plots_class,
    "outliers": outliers_class,
    "profile": profile_class
}

accessors = {**engine_accessors, **dataframe_accessors}

def _create_new_variable(base_name, names):
    while base_name in names:
        base_name = _increment_variable_name(base_name)
        
    return base_name
    
def _increment_variable_name(variable_name):
    match = re.search(r'\d+$', variable_name)
    if match:
        variable_name = variable_name[0:match.start()] + str(int(variable_name[match.start():match.endpos])+1)
    else:
        variable_name = variable_name + "2"
    return variable_name

def _arguments(args, args_properties=None):
    
    args_list = []
    
    # It doesn't find properties or allows kwargs
    if not args_properties or "kwargs" in args_properties:
        args_list = list(args.keys())
        for key in ["source", "target", "operation", "operation_options"]:
            if key in args_list:
                args_list.remove(key)
    
    # Properties from declaration
    else:
        args_list = args_properties.keys()

    # Formating
    for arg in args_list:
        # Variable names of list of variable names
        if arg in args:
            if arg in args_properties and is_any_optimus_type(args_properties[arg].get("type", None)):
                if is_list(args[arg]):
                    args[arg] = f"[{', '.join(args[arg])}]"
                else:
                    args[arg] = args[arg]
            # Native types
            elif is_str(arg) and args_properties[arg].get("type", None) == dict:
                pass
            else:
                args[arg] = pformat(args[arg])

        
    return ", ".join([f'{arg}={args[arg]}' for arg in args_list if arg in args])

def _generate_code_target(body, properties, target):

    arguments = _arguments(body, properties["arguments"])
    
    code = ''

    if target:
        code += f'{target} = '
    
    if body.get("source", None):
        code += f'{body["source"]}.'

    code += f'{body["operation"]}({arguments})'
    
    if target:
        return code, [target]
    else:
        return code, []

def _generate_code_dataframe_transformation(body, properties, variables):

    target = body.get("target")

    if target is None:
        target = body["source"]

    options = body.get("operation_options", None)
    if options:
      if options.get("creates_new", False):
        target = available_variable("df", variables)
    return _generate_code_target(body, properties, target)

def _generate_code_dataframe_mask(body, properties, variables):
    target = body.get("target", available_variable("mask", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_dataframe_clusters(body, properties, variables):    
    target = body.get("target", available_variable("clusters", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_output(body, properties, variables):    
    target = body.get("target", "result")
    return _generate_code_target(body, properties, target)

def _generate_code_engine_dataframe(body, properties, variables):    
    target = body.get("target", available_variable("df", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_engine_connection(body, properties, variables):    
    target = body.get("target", available_variable("conn", variables))
    return _generate_code_target(body, properties, target)

def _generate_code_engine(body, properties, variables):    
    target = body.get("target", available_variable("op", variables))
    return _generate_code_target(body, properties, target)

def _get_generator(func_properties, method_root_type):

    if method_root_type == "dataframe":
    
        # If the method returns a dataframe, it's a transformation
        if DataFrameType == func_properties["return_annotation"]:
            return _generate_code_dataframe_transformation
        # TO-DO: should mask functions be treated as transformations? (MaskDataFrameType)
        if MaskDataFrameType == func_properties["return_annotation"]:
            return _generate_code_dataframe_mask
        # If the method returns anything else, it's an output
        elif ClustersType == func_properties["return_annotation"]:
            return _generate_code_dataframe_clusters
        else:
            return _generate_code_output
    
    elif method_root_type == "engine":
    
        if DataFrameType == func_properties["return_annotation"]:
            return _generate_code_engine_dataframe
        elif ConnectionType == func_properties["return_annotation"]:
            return _generate_code_engine_connection
        else:
            return _generate_code_output
        
    elif method_root_type == "optimus":
        
        return _generate_code_engine
    
    elif method_root_type == "clusters":
    
        return _generate_code_output
        
def _get_method_root_type(accessor):
    if accessor in engine_accessors:
        return "engine"
    if accessor in dataframe_accessors:
        return "dataframe"

def _init_methods(engine):

    method = ""

    if engine == "pandas":
        from optimus.engines.pandas.engine import PandasEngine
        method = PandasEngine
    if engine == "vaex":
        from optimus.engines.vaex.engine import VaexEngine
        method = VaexEngine
    if engine == "spark":
        from optimus.engines.spark.engine import SparkEngine
        method = SparkEngine
    if engine == "dask":
        from optimus.engines.dask.engine import DaskEngine
        method = DaskEngine
    if engine == "ibis":
        from optimus.engines.ibis.engine import IbisEngine
        method = IbisEngine
    if engine == "cudf":
        from optimus.engines.cudf.engine import CUDFEngine
        method = CUDFEngine
    if engine == "dask_cudf":
        from optimus.engines.dask_cudf.engine import DaskCUDFEngine
        method = DaskCUDFEngine
        
    return method

def _generate_code(body=None, variables=[], **kwargs):
    
    if not body:
        body = kwargs

    operation = body["operation"].split(".")

    method = None
    method_root_type = None
    
    if operation[0] == "Optimus":
        method = _init_methods(body["engine"])
        method_root_type = "optimus"
        operation = []
    elif operation[0] in accessors:
        method = accessors[operation[0]]
        method_root_type = _get_method_root_type(operation[0])
        operation = operation[1:]
    elif getattr(dataframe_class, operation[0], None):
        method = dataframe_class
        method_root_type = "dataframe"
    elif getattr(engine_class, operation[0], None):
        method = engine_class
        method_root_type = "engine"
    elif getattr(clusters_class, operation[0], None):
        method = clusters_class
        method_root_type = "clusters"

    for item in operation:
        method = getattr(method, item)

    properties = method_properties(method, method_root_type)

    code, updated = properties["generator"](body, properties, variables)
    
    return code, updated

def optimus_variables():
    from optimus.helpers.functions import engines, dataframes, clusters, connections
    return [ *engines(), *dataframes(), *clusters(), *connections() ]

def available_variable(name, variables):
    return _create_new_variable(name, [*variables, *optimus_variables()])


def method_properties(func, method_root_type):
    
    try:
        fp = signature(func)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}
    except ValueError as e:
        if getattr(func, "funcs"):
            func_properties = {"parameters": {}, "return_annotation": None}
            for f in func.funcs.values():
                fp = signature(f)
                func_properties["parameters"].update(fp.parameters)
                if fp.return_annotation:
                    func_properties["return_annotation"] = fp.return_annotation
            print(func_properties)
        else:
            raise e

    if list(func_properties["parameters"].keys()) == ['root'] and getattr(func, "__call__"):
        fp = signature(func.__call__)
        func_properties = {"parameters": fp.parameters, "return_annotation": fp.return_annotation}

    arguments_list = list(func_properties["parameters"].items())
    arguments = {}
    for key, arg in arguments_list:
        
        if arg.name in ['self', 'cls']:
            continue
        
        arguments[arg.name] = {}
        
        if arg.annotation is not inspect._empty:
            arguments[arg.name].update({"type": arg.annotation})
            
        if arg.default is not inspect._empty:
            arguments[arg.name].update({"value": arg.default})
                        
    return {
        "arguments": arguments, 
        "returns": func_properties["return_annotation"],
        "generator": _get_generator(func_properties, method_root_type)
    }

def generate_code(body=None, variables=[], get_updated=False, **kwargs):

    if not body:
        body = kwargs

    body = val_to_list(body)

    updated = []

    code = []

    for operation in body:
        
        operation = prepare(operation)
        operation_code, operation_updated = _generate_code(operation, [*updated, *variables])
        updated.extend(operation_updated)
        code.append(operation_code)

    if get_updated:
        return "\n".join(code), one_list_to_val(updated)
    else:
        return "\n".join(code)
    



from dask import dataframe as dd

from optimus.infer import is_dict, is_dict_of_one_element, is_list_value, is_list_of_one_element


def tuple_to_dict(value):
    """
    Convert tuple to dict
    :param value: tuple to be converted
    :return:
    """

    return dict((x, y) for x, y in value)


def format_dict(_dict, tidy=True):
    """
    This function format a dict. If the main dict or a deep dict has only on element
     {"col_name":{0.5: 200}} we get 200
    :param _dict: dict to be formatted
    :param tidy:
    :return:
    """

    if tidy is True:
        def _format_dict(_dict):

            if not is_dict(_dict):
                return _dict
            for k, v in _dict.items():
                # If the value is a dict
                if is_dict(v):
                    # and only have one value
                    if len(v) == 1:
                        _dict[k] = next(iter(v.values()))
                else:
                    if len(_dict) == 1:
                        _dict = v
            return _dict

        if is_list_of_one_element(_dict):
            _dict = _dict[0]
        elif is_dict_of_one_element(_dict):
            # if dict_depth(_dict) >4:
            _dict = next(iter(_dict.values()))

        # Some aggregation like min or max return a string column
        def repeat(f, n, _dict):
            if n == 1:  # note 1, not 0
                return f(_dict)
            else:
                return f(repeat(f, n - 1, _dict))  # call f with returned value

        # TODO: Maybe this can be done in a recursive way
        # We apply two passes to the dict so we can process internals dicts and the superiors ones
        return repeat(_format_dict, 2, _dict)
    else:
        # Return the dict from a list
        if is_list_value(_dict):
            return _dict[0]
        else:
            return _dict


#
# def cudf_series_to_pandas(serie):
#     return serie.to_pandas()
#
#
# def dask_dataframe_to_dask_cudf(df):
#     import cudf
#     return df.map_partitions(cudf.DataFrame.from_pandas)
#
#
# # To cudf
# def dask_dataframe_to_cudf(df):
#     return pandas_to_cudf(dask_dataframe_to_pandas(df))
#
#
# def dask_cudf_to_cudf(df):
#     return df.compute()


# To Pandas
def spark_to_pandas(df):
    return df.toPandas()


def dask_cudf_to_pandas(df):
    return df.map_partitions(lambda df: df.to_pandas())


def dask_dataframe_to_pandas(df):
    return df.compute()


def cudf_to_pandas(df):
    return df.to_pandas()


#
def cudf_to_dask_cudf(df, n_partitions=1):
    import dask_cudf
    return dask_cudf.from_cudf(df, npartitions=1)

#
#
# def cudf_to_cupy_arr(df):
#     import cupy as cp
#     return cp.fromDlpack(df.to_dlpack())
#
#
# def pandas_to_cudf(df):
#     import cudf
#     return cudf.from_pandas(df)


def pandas_to_dask_dataframe(pdf, n_partitions=1):
    return dd.from_pandas(pdf, npartitions=n_partitions)


def pandas_to_vaex_dataframe(pdf, n_partitions=1):
    import vaex
    return vaex.from_pandas(pdf)


def pandas_to_dask_cudf_dataframe(pdf, n_partitions=1):
    import cudf
    import dask_cudf
    # Seems that from_cudf also accepts pandas
    cdf = cudf.DataFrame.from_pandas(pdf)
    return dask_cudf.from_cudf(cdf, npartitions=n_partitions)



import operator
from optimus.engines.base.set import BaseSet
from optimus.helpers.types import DataFrameType, InternalDataFrameType
import time
from abc import abstractmethod, ABC

import humanize
import imgkit
import jinja2
import simplejson as json
from dask import dataframe as dd
from glom import assign
from tabulate import tabulate

from optimus.engines.base.stringclustering import string_clustering
from optimus.helpers.check import is_notebook
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import BUFFER_SIZE, Actions, ProfilerDataTypes, RELATIVE_ERROR
from optimus.helpers.core import val_to_list
from optimus.helpers.functions import absolute_path, reduce_mem_usage, update_dict
from optimus.helpers.json import json_converter
from optimus.helpers.output import print_html
from optimus.infer import is_str, is_tuple, is_list
from optimus.profiler.constants import MAX_BUCKETS
from optimus.profiler.templates.html import HEADER, FOOTER
from .columns import BaseColumns
from .meta import Meta
from .profile import BaseProfile
from ...outliers.outliers import Outliers
from ...plots.functions import plot_hist, plot_frequency
from ...plots.plots import Plot


class BaseDataFrame(ABC):
    """
    Optimus DataFrame
    """

    def __init__(self, data):
        self.data = data
        self.buffer = None
        self.updated = None
        self.meta = {}

    def __del__(self):
        del self.data

    @property
    def root(self):
        return self

    def _repr_html_(self):
        df = self
        return df.table()

    def __repr__(self):
        df = self
        return df.ascii()

    def __getitem__(self, item):

        if isinstance(item, slice):
            return self.buffer_window("*", item.start, item.stop)
        elif is_str(item) or is_list(item):
            return self.cols.select(item)
        elif isinstance(item, BaseDataFrame):
            return self.rows.select(item)

    def __setitem__(self, key, value):

        df = self.cols.assign({key: value})
        self.data = df.data
        self.buffer = df.buffer
        self.updated = df.updated
        self.meta = df.meta

    def __len__(self):
        return self.rows.count()

    def new(self, dfd, meta=None) -> DataFrameType:
        df = self.__class__(dfd)
        if meta is not None:
            df.meta = meta
        return df

    @staticmethod
    def __operator__(df, dtype, multiple_columns=False) -> DataFrameType:
        if isinstance(df, (BaseDataFrame,)):
            col1 = "*" if multiple_columns else df.cols.names(0)[0]

            if dtype:
                df = df.cols.cast(col1, dtype).data
            else:
                df = df.data

            if not multiple_columns:
                df = df[col1]

        return df

    @abstractmethod
    def _base_to_dfd(self, pdf, n_partitions) -> InternalDataFrameType:
        pass

    @abstractmethod
    def to_optimus_pandas(self) -> DataFrameType:
        pass

    def unary_operation(self, df, opb, dtype=None) -> DataFrameType:
        """
        Helper to process binary operations
        :param df: Left
        :param opb: Operator
        :param dtype: 
        :return:
        """
        df = BaseDataFrame.__operator__(df, dtype, True)

        return self.new(opb(df))

    def operation(self, df1, df2, opb, dtype=None) -> DataFrameType:
        """
        Helper to process binary operations
        :param df1: Left
        :param df2: Right
        :param opb: Operator
        :param dtype:
        :return:
        """
        if not isinstance(df1, (BaseDataFrame,)) or not isinstance(df2, (BaseDataFrame,)):
            multiple_columns = True
        else:
            multiple_columns = df1.cols.names() == df2.cols.names()
        df1 = BaseDataFrame.__operator__(df1, dtype, multiple_columns)
        df2 = BaseDataFrame.__operator__(df2, dtype, multiple_columns)

        if multiple_columns:
            df = self.new(opb(df1, df2))

        else:

            name_left = name_right = ""

            if not isinstance(df1, (int, float, str, dict, list)):
                name_left = getattr(df1, "name", 0)

            if not isinstance(df2, (int, float, str, dict, list)):
                name_right = getattr(df2, "name", 0)

            if name_left and name_right:
                name = (name_left + "_" + name_right) if name_left != name_right else name_left
            else:
                name = name_left if name_left else name_right

            df = self.new(opb(df1, df2).rename(name).to_frame())

        return df

    def __invert__(self) -> DataFrameType:
        return self.unary_operation(self, operator.invert)

    def __neg__(self) -> DataFrameType:
        return self.unary_operation(self, operator.neg)

    def __add__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.add)

    def __radd__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.add)

    def __sub__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.sub, "float")

    def __rsub__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.sub, "float")

    def __mul__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.mul, "float")

    def __rmul__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.mul, "float")

    def __truediv__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.truediv, "float")

    def __rtruediv__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.truediv, "float")

    def __floordiv__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.floordiv, "float")

    def __rfloordiv__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.floordiv, "float")

    def __mod__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.mod, "float")

    def __rmod__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.mod, "float")

    def __pow__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.pow, "float")

    def __rpow__(self, df2) -> DataFrameType:
        return self.operation(df2, self, operator.pow, "float")

    def __eq__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.eq)

    def __gt__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.gt, "float")

    def __lt__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.lt, "float")

    def __ne__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.ne)

    def __ge__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.ge, "float")

    def __le__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.le, "float")

    def __and__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.__and__, "bool")

    def __or__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.__or__, "bool")

    def __xor__(self, df2) -> DataFrameType:
        return self.operation(self, df2, operator.__xor__, "bool")

    def _to_values(self):
        """
        Return values from a dataframe in numpy or cupy format. Aimed to be used internally in Machine Learning models
        :return:
        """
        return self.data.values

    def equals(self, df2: DataFrameType) -> bool:
        if isinstance(df2, (BaseDataFrame,)):
            df2 = df2.data
            return self.data.equals(df2)
        else:
            return self.to_dict(n="all") == df2

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def rows(self):
        pass

    def cols(self):
        return BaseColumns

    @property
    def profile(self):
        return BaseProfile(self)

    @property
    def set(self):
        return BaseSet(self)

    @property
    def plot(self):
        return Plot(self)

    @property
    def outliers(self):
        return Outliers(self)

    @abstractmethod
    def encoding(self):
        pass

    @abstractmethod
    def visualize(self):
        pass

    @staticmethod
    @abstractmethod
    def execute():
        pass

    @staticmethod
    @abstractmethod
    def compute():
        # We will handle all dataframe as if the could compute the result,
        # something that only can be done in dask and in spark triggering and action.
        # With this we expect to abstract the behavior and just use compute() a value from operation
        pass

    def _assign(self, kw_columns):

        dfd = self.data
        return dfd.assign(**kw_columns)

    def to_json(self, cols="*", n="all", orient="list"):
        """
        Return a json from a Dataframe
        :return:
        """

        return json.dumps(self.to_dict(cols, n, orient), ensure_ascii=False, default=json_converter)

    def to_dict(self, cols="*", n=10, orient="list"):
        """
            Return a dict from a Collect result
            :param n:
            :param orient:
            :return:
        """
        if n=="all":
            dfd = self.cols.select(cols).to_pandas()
        else:
            dfd = self.buffer_window(cols, 0, n).data
            
        return dfd.to_dict(orient)

    @staticmethod
    @abstractmethod
    def sample(n=10, random=False):
        pass

    def columns_sample(self, columns="*"):
        """
        Return a dict of the sample of a Dataframe
        :return:
        """
        df = self

        return {"columns": [{"title": col_name} for col_name in df.cols.select(columns).cols.names()],
                "value": df.rows.to_list(columns)}

    @abstractmethod
    def to_pandas(self):
        pass

    def stratified_sample(self, col_name, seed: int = 1):
        """
        Stratified Sampling
        columns_type = parse_columns(df, columns_type.keys())
        :param col_name:
        :param seed:
        :return:
        """
        df = self.data
        # n = min(5, df[col_name].value_counts().min())
        df = df.groupby(col_name).apply(lambda x: x.sample(2))
        # df_.index = df_.index.droplevel(0)
        return df

    def get_buffer(self):
        return self.buffer

    @abstractmethod
    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        pass

    def _reset_buffer(self):
        self.buffer = None

    def reset_buffer(self):
        self._reset_buffer()
        self.meta = Meta.reset(self.meta, "buffer_time")

    def buffer_window(self, columns="*", lower_bound=None, upper_bound=None, n=BUFFER_SIZE):

        df = self

        meta = df.meta
        buffer_time = Meta.get(meta, "buffer_time")
        last_action_time = Meta.get(meta, "last_action_time")

        if buffer_time and last_action_time:
            if buffer_time > last_action_time:
                self.set_buffer(columns, n)

        if lower_bound is None or lower_bound < 0:
            lower_bound = 0

        input_columns = parse_columns(df, columns)
        return self._buffer_window(input_columns, lower_bound, upper_bound)

    def buffer_json(self, columns):
        df = self.buffer
        columns = parse_columns(df, columns)

        return {"columns": [{"title": col_name} for col_name in df.cols.select(columns).cols.names()],
                "value": df.rows.to_list(columns)}

    def size(self, deep=True, format=None):
        """
        Get the size of a dask in bytes
        :return:
        """
        df = self.data
        result = df.memory_usage(index=True, deep=deep).sum()
        if format == "human":
            result = humanize.naturalsize(result)

        return result

    def optimize(self, categorical_threshold=50, verbose=False):
        df = self
        return reduce_mem_usage(df, categorical_threshold=categorical_threshold, verbose=verbose)

    def run(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :return:
        """
        df = self
        df.cache().count()
        return df

    @staticmethod
    @abstractmethod
    def query(sql_expression):
        raise NotImplementedError

    @staticmethod
    def is_cached(df):
        """

        :return:
        """
        return False if Meta.get(df.meta, "profile") is None else True

    def _cols_to_profile(self, columns):
        """
        Get the columns that needs to be profiled and renames the columns in the metadata
        :return:
        """

        df = self
        actions = Meta.get(df.meta, "transformations.actions")
        has_actions = actions is not None and len(actions) > 0

        profiler_columns = Meta.get(df.meta, "profile.columns")

        new_columns = parse_columns(df, columns)

        if profiler_columns is None:
            calculate_columns = new_columns
        else:
            profiled_columns = list(profiler_columns.keys())
            if not has_actions:
                calculate_columns = [column for column in new_columns if column not in profiled_columns]

            else:
                modified_columns = []
                # Operations need to be processed int the same order that created
                for l in Meta.get(df.meta, "transformations.actions"):
                    dropped_columns = []
                    for action_name, column in l.items():
                        if is_tuple(column):
                            source, target = column

                        elif is_str(column):
                            source = target = column

                        if action_name == Actions.COPY.value and source in profiler_columns:
                            profiler_columns[target] = profiler_columns[source]

                        elif action_name == Actions.RENAME.value and source in profiler_columns:
                            profiler_columns[target] = profiler_columns[source]
                            profiler_columns.pop(source)
                            if source in new_columns:
                                new_columns.pop(source)
                                new_columns.push(target)

                        elif action_name == Actions.DROP.value and source in profiler_columns:
                            profiler_columns.pop(source)
                            dropped_columns.append(source)

                        else:
                            modified_columns.append(source)

                profiled_columns = list(profiler_columns.keys())

                calculate_columns = [column for column in new_columns if column not in profiled_columns]
                modified_columns = [column for column in modified_columns if column in new_columns]
                calculate_columns = list(set(modified_columns + calculate_columns))
                calculate_columns = list(set(calculate_columns) - set(dropped_columns))

        return calculate_columns

    @abstractmethod
    def partitions(self):
        pass

    @staticmethod
    def partitioner():
        raise NotImplementedError

    def repartition(self, n=None, *args, **kwargs):
        df = self.data
        return self.new(df, meta=self.meta)

    def table_image(self, path, limit=10):
        """
        Output table as image
        :param limit:
        :param path:
        :return:
        """

        css = absolute_path("/css/styles.css")

        imgkit.from_string(self.table_html(limit=limit, full=True), path, css=css)
        print_html("<img src='" + path + "'>")

    def table_html(self, limit=10, columns=None, title=None, full=False, truncate=True, count=True, highlight=[]):
        """
        Return a HTML table with the spark cols, data types and values
        :param columns: Columns to be printed
        :param limit: How many rows will be printed
        :param title: Table title
        :param full: Include html header and footer
        :param truncate: Truncate the row information
        :param count:

        :return:
        """

        columns = parse_columns(self, columns)
        if limit is None:
            limit = 10

        df = self

        total_rows = df.rows.approx_count()

        if limit == "all":
            limit = total_rows
            data = df.cols.select(columns).to_dict(n="all", orient="records")
        else:
            limit = min(limit, total_rows)
            data = df.cols.select(columns).rows.limit(limit + 1).to_dict(n="all", orient="records")
        # Load the Jinja template
        template_loader = jinja2.FileSystemLoader(searchpath=absolute_path("/templates/out"))
        template_env = jinja2.Environment(loader=template_loader, autoescape=True)
        template = template_env.get_template("table.html")

        # Filter only the columns and data type info need it
        dtypes = [(k, v) for k, v in df.cols.dtypes().items()]

        # Remove not selected columns
        final_columns = []
        for i in dtypes:
            for j in columns:
                if i[0] == j:
                    final_columns.append(i)

        total_rows = humanize.intword(total_rows)
        total_cols = df.cols.count()
        total_partitions = df.partitions()
        df_type = type(df)
        highlight = val_to_list(highlight)
        output = template.render(df_type=df_type, cols=final_columns, data=data, limit=limit, total_rows=total_rows,
                                 total_cols=total_cols,
                                 partitions=total_partitions, title=title, truncate=truncate, highlight=highlight)

        if full is True:
            output = HEADER + output + FOOTER
        return output

    def display(self, limit=10, columns=None, title=None, truncate=True, plain_text=False, highlight=[]):
        # TODO: limit, columns, title, truncate
        df = self

        if is_notebook() and not plain_text:
            print_html(df.table(limit, columns, title, truncate, highlight))

        else:
            print(df.ascii(limit, columns))

    def print(self, limit=10, columns=None):
        print(self.ascii(limit, columns))

    def table(self, limit=None, columns=None, title=None, truncate=True, highlight=[]):
        df = self
        try:
            if is_notebook():
                # TODO: move the html param to the ::: if is_notebook() and engine.output is "html":
                return df.table_html(title=title, limit=limit, columns=columns, truncate=truncate, highlight=highlight)

        except NameError as e:
            print(e)

        return df.ascii(limit, columns)

    def ascii(self, limit=10, columns=None):
        df = self
        if not columns:
            columns = "*"

        limit = min(limit, df.rows.approx_count())
        return tabulate(df.rows.limit(limit + 1).cols.select(columns).to_pandas(),
                        headers=[f"""{i}\n({j})""" for i, j in df.cols.dtypes().items()],
                        tablefmt="simple",
                        showindex="never")+"\n"

    def export(self):
        """
        Helper function to export all the dataframe in dict format. Aimed to be used in test functions
        :return:
        """
        df_dict = self.to_dict(n="all")
        df_schema = self.cols.dtypes() # TO-DO use types in tests

        df_data = []

        from pprint import pformat

        for col_name in df_dict.keys():
            value = pformat((col_name, df_schema[col_name]))
            value += ": "
            value += json.dumps(df_dict[col_name], ensure_ascii=False, default=json_converter)
            df_data.append(value)

        return "{" + ", ".join(df_data) + "}"

    def show(self, n=10):
        """
        :return:
        """
        return self.data.head(n=n)

    @staticmethod
    @abstractmethod
    def debug():
        pass

    def reset(self):
        # df = self.df
        df = self
        df.meta = {}
        return df

    def calculate_profile(self, columns="*", bins: int = MAX_BUCKETS, flush: bool = False, size=False):
        """
        Returns a new dataframe with the profile data in its added to the meta property
        :param columns:
        :param bins:
        :param flush:
        :param size: get the dataframe size in memory. Use with caution this could be slow for big data frames.
        :return:
        """
        _t = time.process_time()
        profiler_time = {"hist": {}, "frequency": {}, "count_mismatch": {}}

        df = self
        meta = self.meta

        if flush is False:
            cols_to_profile = df._cols_to_profile(columns) or []
        else:
            cols_to_profile = parse_columns(df, columns) or []

        profiler_data = Meta.get(meta, "profile")

        is_cached = profiler_data is not None

        if profiler_data is None:
            profiler_data = {}
        cols_dtypes = None

        profiler_time["beginning"] = {"elapsed_time": time.process_time() - _t}
        if cols_to_profile or not is_cached or flush is True:
            # Reset profiler metadata
            meta = Meta.set(meta, "profile", {})
            df.meta = meta

            hist_cols = []
            freq_cols = []

            cols_dtypes = {}
            cols_to_infer = [*cols_to_profile]

            for col_name in cols_to_profile:
                _props = Meta.get(df.meta, f"columns_dtypes.{col_name}")

                if _props is not None:
                    cols_dtypes[col_name] = _props
                    cols_to_infer.remove(col_name)

            if cols_to_infer:
                cols_dtypes = {**cols_dtypes, **df.cols.infer_profiler_dtypes(cols_to_infer)}
                cols_dtypes = {col: cols_dtypes[col] for col in cols_to_profile}

            _t = time.process_time()
            mismatch = df.cols.count_mismatch(cols_dtypes)
            profiler_time["count_mismatch"] = {"columns": cols_dtypes, "elapsed_time": time.process_time() - _t}

            # Get with columns are numerical and does not have mismatch so we can calculate the histogram
            cols_properties = cols_dtypes.items()
            for col_name, properties in cols_properties:
                if properties.get("categorical") is True \
                        or properties.get("dtype") == ProfilerDataTypes.EMAIL.value \
                        or properties.get("dtype") == ProfilerDataTypes.URL.value \
                        or properties.get("dtype") == ProfilerDataTypes.OBJECT.value:
                    freq_cols.append(col_name)
                else:
                    hist_cols.append(col_name)

            hist = None
            freq = {}
            sliced_freq = {}
            count_uniques = None

            if len(hist_cols):
                _t = time.process_time()
                hist = df.cols.hist(hist_cols, buckets=bins, compute=False)
                profiler_time["hist"] = {"columns": hist_cols, "elapsed_time": time.process_time() - _t}

            if len(freq_cols):
                _t = time.process_time()
                sliced_cols = []
                non_sliced_cols = []

                # Extract the columns with cells larger thatn
                max_cell_length = getattr(df.meta, "max_cell_length", None)

                if max_cell_length:
                    for i, j in max_cell_length.items():
                        if i in freq_cols:
                            if j > 50:
                                sliced_cols.append(i)
                            else:
                                non_sliced_cols.append(i)

                else:
                    non_sliced_cols = freq_cols

                if len(non_sliced_cols) > 0:
                    # print("non_sliced_cols",non_sliced_cols)
                    freq = df.cols.frequency(non_sliced_cols, n=bins, count_uniques=True, compute=False)

                if len(sliced_cols) > 0:
                    # print("sliced_cols", sliced_cols)
                    sliced_freq = df.cols.slice(sliced_cols, 0, 50).cols.frequency(sliced_cols, n=bins,
                                                                                   count_uniques=True,
                                                                                   compute=False)

                profiler_time["frequency"] = {"columns": freq_cols, "elapsed_time": time.process_time() - _t}

            def merge(_columns, _hist, _freq, _mismatch, _dtypes, _count_uniques):
                _c = {}

                _hist = {} if _hist is None else _hist.get("hist", {})
                _freq = {} if _freq is None else _freq.get("frequency", {})

                for _col_name in _columns:
                    _c[_col_name] = {"stats": _mismatch[_col_name], "dtype": _dtypes[_col_name]}
                    if _col_name in _freq:
                        f = _freq[_col_name]
                        _c[_col_name]["stats"]["frequency"] = f["values"]
                        _c[_col_name]["stats"]["count_uniques"] = f["count_uniques"]

                    elif _col_name in _hist:
                        h = _hist[_col_name]
                        _c[_col_name]["stats"]["hist"] = h

                return {"columns": _c}

            # Nulls
            total_count_na = 0

            dtypes = df.cols.dtypes("*")

            hist, freq, sliced_freq, mismatch = dd.compute(hist, freq, sliced_freq, mismatch)

            freq = {**freq, **sliced_freq}

            updated_columns = merge(cols_to_profile, hist, freq, mismatch, dtypes, count_uniques)
            profiler_data = update_dict(profiler_data, updated_columns)

            assign(profiler_data, "name", Meta.get(df.meta, "name"), dict)
            assign(profiler_data, "file_name", Meta.get(df.meta, "file_name"), dict)

            data_set_info = {'cols_count': df.cols.count(),
                             'rows_count': df.rows.count(),
                             }
            if size is True:
                data_set_info.update({'size': df.size(format="human")})

            assign(profiler_data, "summary", data_set_info, dict)
            dtypes_list = list(set(df.cols.dtypes("*").values()))
            assign(profiler_data, "summary.dtypes_list", dtypes_list, dict)
            assign(profiler_data, "summary.total_count_dtypes", len(set([i for i in dtypes.values()])), dict)
            assign(profiler_data, "summary.missing_count", total_count_na, dict)
            assign(profiler_data, "summary.p_missing", round(total_count_na / df.rows.count() * 100, 2))

        # _t = time.process_time()

        all_columns_names = df.cols.names()

        meta = Meta.set(meta, "transformations", value={})

        # Order columns
        actual_columns = profiler_data["columns"]
        profiler_data["columns"] = {key: actual_columns[key] for key in all_columns_names if key in actual_columns}
        meta = Meta.set(meta, "profile", profiler_data)

        if cols_dtypes is not None:
            df.meta = meta
            df = df.cols.set_dtype(cols_dtypes, True)
            meta = df.meta

        # Reset Actions
        meta = Meta.reset_actions(meta)
        df.meta = meta
        profiler_time["end"] = {"elapsed_time": time.process_time() - _t}
        # print(profiler_time)
        return df

    def graph(self) -> dict:
        """
        Return a dict the Dask tasks graph
        :return:
        """
        dfd = self.data

        return dfd.__dask_graph__().layers

    def get_series(self):
        col1 = self.cols.names(0)[0]
        return self.data[col1]

    def string_clustering(self, columns="*", algorithm="fingerprint", *args, **kwargs):
        return string_clustering(self, columns, algorithm, *args, **kwargs)
        # return clusters

    def agg(self, aggregations: dict, groupby=None, output="dict"):

        df = self
        dfd = df.data

        if groupby:
            groupby = parse_columns(df, groupby)

            for column, aggregations_set in aggregations.items():
                aggregations[column] = val_to_list(aggregations_set)

            dfd = dfd.groupby(groupby).agg(aggregations)

            dfd.columns = ['_'.join(col).strip() for col in dfd.columns.values]

            if output == "dict":
                result = dfd.to_dict()

            elif output == "dataframe":
                result = self.new(dfd.reset_index())

        else:
            result = {}

            for column, aggregations_set in aggregations.items():
                aggregations_set = val_to_list(aggregations_set)
                for aggregation in aggregations_set:
                    result[column + "_" + aggregation] = getattr(df.cols, aggregation)(column, tidy=True)

            if output == "dataframe":
                result = self.new(result)

        return result

    def report(self, df, columns="*", buckets=MAX_BUCKETS, infer=False, relative_error=RELATIVE_ERROR,
               approx_count=True,
               mismatch=None, advanced_stats=True):
        """
        Return dataframe statistical information in HTML Format
        :param df: Dataframe to be analyzed
        :param columns: Columns to be analyzed
        :param buckets: Number of buckets calculated to print the histogram
        :param infer: infer data type
        :param relative_error: Relative Error for quantile discretizer calculation
        :param approx_count: Use approx_count_distinct or countDistinct
        :param mismatch:
        :param advanced_stats:
        :return:
        """

        columns = parse_columns(df, columns)
        output = self.dataset(df, columns, buckets, infer, relative_error, approx_count, format="dict",
                              mismatch=mismatch, advanced_stats=advanced_stats)

        # Load jinja
        template_loader = jinja2.FileSystemLoader(searchpath=absolute_path("/profiler/templates/out"))
        template_env = jinja2.Environment(loader=template_loader, autoescape=True)

        # Render template
        # Create the profiler info header
        html = ""
        general_template = template_env.get_template("general_info.html")
        html = html + general_template.render(data=output)

        template = template_env.get_template("one_column.html")
        # Create every column stats
        for col_name in columns:
            hist_pic = None
            freq_pic = None

            col = output["columns"][col_name]
            if "hist" in col["stats"]:
                hist_dict = col["stats"]["hist"]

                if col["column_dtype"] == "date":
                    hist_year = plot_hist({col_name: hist_dict["years"]}, "base64", "years")
                    hist_month = plot_hist({col_name: hist_dict["months"]}, "base64", "months")
                    hist_weekday = plot_hist({col_name: hist_dict["weekdays"]}, "base64", "weekdays")
                    hist_hour = plot_hist({col_name: hist_dict["hours"]}, "base64", "hours")
                    hist_minute = plot_hist({col_name: hist_dict["minutes"]}, "base64", "minutes")
                    hist_pic = {"hist_years": hist_year, "hist_months": hist_month, "hist_weekdays": hist_weekday,
                                "hist_hours": hist_hour, "hist_minutes": hist_minute}

                elif col["column_dtype"] == "int" or col["column_dtype"] == "string" or col[
                    "column_dtype"] == "decimal":
                    hist = plot_hist({col_name: hist_dict}, output="base64")
                    hist_pic = {"hist_numeric_string": hist}
            if "frequency" in col:
                freq_pic = plot_frequency({col_name: col["frequency"]}, output="base64")

            html = html + template.render(data=col, freq_pic=freq_pic, hist_pic=hist_pic)

        # Save in case we want to output to a html file
        # self.html = html + df.table_html(10)
        self.html = html

        # Display HTML
        print_html(self.html)

        # JSON
        # Save in case we want to output to a json file
        self.json = output

        return self



from abc import abstractmethod

import dask
import humanize
from dask.distributed import Variable
from dask.utils import parse_bytes

from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.functions import random_int
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_one_element


class DaskBaseDataFrame(BaseDataFrame):

    def __init__(self, data):
        super().__init__(data)

    def _assign(self, kw_columns):

        dfd = self.root.data

        if dfd.known_divisions:
            for key in kw_columns:
                kw_column = kw_columns[key]
                if not is_one_element(kw_column) and not callable(kw_column) and not getattr(kw_column, "known_divisions", None):
                    _dfd = kw_column.reset_index().set_index('index')
                    if key in _dfd:
                        # the incoming series has the same column key
                        kw_columns[key] = _dfd[key]
                    else:
                        # the incoming series has no column key
                        kw_columns[key] = _dfd[0]
        return dfd.assign(**kw_columns)

    @staticmethod
    @abstractmethod
    def _base_to_dfd(df, n_partitions):
        """
        Convert a dataframe from the dataframe base (pandas, cudf) in a distributed engine dataframe (dask, dask_cudf)
        :param pdf:
        :param n_partitions:
        :return:
        """
        pass

    def execute(self):
        self.data = self.data.persist()
        return self

    def compute(self):
        df = self.data
        return df.compute()

    def visualize(self):
        return display(self.data.visualize())

    def export(self):
        """
        Helper function to export all the dataframe in text format. Aimed to be used in test functions
        :return:
        """
        df = self.root
        df_data = df.to_json()
        df_schema = df.cols.dtypes()

        return f"{df_schema}, {df_data}"

    def _reset_buffer(self):
        if self.buffer:
            Variable(self.buffer).delete()
            self.buffer = None

    def get_buffer(self):
        return Variable(self.buffer).get() if self.buffer else None

    def _buffer_window(self, input_cols, lower_bound, upper_bound):
        def func(value):
            return value[lower_bound:upper_bound]

        return PandasDataFrame(self.data[input_cols].partitions[0].map_partitions(func).compute())

    def sample(self, n=10, random=False):
        """
        Return a n number of sample from a dataFrame
        :param n: Number of samples
        :param random: if true get a semi random sample
        :return:
        """
        df = self.root
        if random is True:
            seed = random_int()
        elif random is False:
            seed = 0
        else:
            RaiseIt.value_error(random, ["True", "False"])

        rows_count = df.rows.count()
        if n < rows_count:
            # n/rows_count can return a number that represent less the total number we expect. multiply by 1.1
            fraction = (n / rows_count) * 1.1
        else:
            fraction = 1.0
        return self.root.new(df.data.sample(frac=fraction, random_state=seed))

    def stratified_sample(self, col_name, seed: int = 1):
        """
        Stratified Sampling
        :param col_name:
        :param seed:
        :return:
        """
        df = self.data
        n = min(5, df[col_name].value_counts().min())
        df = df.groupby(col_name).apply(lambda x: x.sample(2))
        # df_.index = df_.index.droplevel(0)
        return self.root.new(df)

    @staticmethod
    def pivot(index, column, values):
        """
        Return reshaped DataFrame organized by given index / column values.
        :param index: Column to use to make new frame's index.
        :param column: Column to use to make new frame's columns.
        :param values: Column(s) to use for populating new frame's values.
        :return:
        """
        raise NotImplementedError

    @staticmethod
    def melt(id_vars, value_vars, var_name="variable", value_name="value", data_type="str"):
        """
        Convert DataFrame from wide to long format.
        :param id_vars: column with unique values
        :param value_vars: Column names that are going to be converted to columns values
        :param var_name: Column name for vars
        :param value_name: Column name for values
        :param data_type: All columns must have the same type. It will transform all columns to this data type.
        :return:
        """

        raise NotImplementedError

    def size(self, deep=False, format=None):
        """
        Get the size of a dask in bytes
        :return:
        """
        df = self.data
        result = df.memory_usage(index=True, deep=deep).sum().compute()
        if format == "human":
            result = humanize.naturalsize(result)

        return result

    def run(self):
        """
        This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that Apache Spark save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        :return:
        """
        df = self.data
        df.cache().count()
        return df

    @staticmethod
    def query(sql_expression):
        raise NotImplementedError

    def partitions(self):
        return self.data.npartitions

    def equals(self, df2) -> bool:
        if isinstance(df2, (BaseDataFrame,)):
            return df2.to_dict(n="all") == self.to_dict(n="all")
        else:
            return self.to_dict(n="all") == df2

    @staticmethod
    def partitioner():
        print("Dask not support custom partitioner")
        raise NotImplementedError

    def repartition(self, n=None, *args, **kwargs):
        dfd = self.data
        df = self
        if n == "auto":
            # Follow a heuristic for partitioning a mentioned
            # https://docs.dask.org/en/latest/best-practices.html#avoid-very-large-partitions
            client = dask.distributed.get_client()
            worker_memory = parse_bytes(client.cluster.worker_spec[0]["options"]["memory_limit"])
            nthreads = client.cluster.worker_spec[0]["options"]["nthreads"]

            part_recommended_size = worker_memory / nthreads / 10
            n = int(df.size() / part_recommended_size)

            # Partition can not be lower than 1
            n = n if n < 0 else 1
            # TODO .repartition(partition_size="100MB"). https://stackoverflow.com/questions/44657631/strategy-for-partitioning-dask-dataframes-efficiently
        dfd = dfd.repartition(npartitions=n, *args, **kwargs)

        return self.new(dfd, meta=self.meta)

    @staticmethod
    def debug():
        """

        :return:
        """
        raise NotImplementedError

    @staticmethod
    def create_id(column="id"):
        """
        Create a unique id for every row.
        :param column: Columns to be processed
        :return:
        """

        raise NotImplementedError

    def to_dict(self, cols="*", n=10, orient="list"):
        """
        Create a dict
        :param n:
        :param orient:
        :return:
        """

        if n=="all":
            series = self.cols.select(cols).to_pandas()
        else:
            series = self.buffer_window(cols, 0, n).data
            
        return series.to_dict(orient)

    def to_pandas(self):
        return self.data.compute()

    @property
    def constants(self):
        from optimus.engines.base.dask.constants import Constants
        return Constants()



from abc import abstractmethod, ABC
from typing import Tuple
from optimus.helpers.types import DataFrameType, DataFrameTypeList, InternalDataFrameType

from multipledispatch import dispatch

from optimus.engines.base.meta import Meta
# This implementation works for Spark, Dask, dask_cudf
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import one_list_to_val, val_to_list
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_str, is_list_of_str_or_int


class BaseRows(ABC):
    """Base class for all Rows implementations"""

    def __init__(self, root):
        self.root = root

    @staticmethod
    @abstractmethod
    def _sort(df, col_name, ascending) -> DataFrameType:
        pass

    def _sort_multiple(self, dfd, meta, col_sort) -> Tuple[InternalDataFrameType, dict]:
        """
        Sort rows taking into account multiple columns
        :param col_sort: column and sort type combination (col_name, "asc")
        :type col_sort: list of tuples
        """

        for cs in col_sort:
            col_name = one_list_to_val(cs[0])
            order = cs[1]

            if order != "asc" and order != "desc":
                RaiseIt.value_error(order, ["asc", "desc"])

            dfd = self._sort(dfd, col_name, True if order == "asc" else False)
            meta = Meta.action(meta, Actions.SORT_ROW.value, col_name)

        return dfd, meta

    def _reverse(self, dfd) -> InternalDataFrameType:
        return dfd[::-1]

    @staticmethod
    @abstractmethod
    def create_id(column="id"):
        pass

    @abstractmethod
    def append(self, dfs: DataFrameTypeList, cols_map) -> DataFrameType:
        pass

    def apply(self, func, args=None, output_cols=None) -> DataFrameType:
        """
        This will aimed to handle vectorized and not vectorized operations
        :param output_cols:
        :param func:
        :return:
        """
        df = self.root
        dfd = self.root.data
        kw_columns = {}

        for output_col in output_cols:
            result = func(dfd, *args)
            kw_columns = {output_col: result}

        return df.cols.assign(kw_columns)

    def select(self, expr=None, contains=None, case=None, flags=0, na=False, regex=False) -> DataFrameType:
        """
        :param expr: Expression used, For Ex: (df["A"] > 3) & (df["A"] <= 1000) or Column name "A"
        :param contains: List of string
        :param case:
        :param flags:
        :param na:
        :param regex:
        :return:
        """

        df = self.root
        dfd = df.data

        if is_str(expr):
            if expr in df.cols.names():
                if contains is not None:
                    expr = df.mask.contains(expr, value=contains, case=case, flags=flags, na=na, regex=regex)
                else:
                    expr = df[expr]
            else:
                expr = eval(expr)
        elif expr:
            expr = expr.get_series()
        dfd = dfd[expr]
        meta = Meta.action(df.meta, Actions.SELECT_ROW.value, df.cols.names())

        df = self.root.new(dfd, meta=meta)
        return df

    def _count(self, compute=True) -> int:
        return len(self.root.data.index)

    def count(self, compute=True) -> int:
        """
        Count dataframe rows
        """
        df = self.root
        dfd = df.data
        # TODO: Be sure that we need the compute param
        if compute is True:
            result = self._count(compute)
        else:
            result = df.functions.delayed(len)(dfd)
        return result

    def to_list(self, input_cols) -> list:
        """

        :param input_cols:
        :return:
        """
        df = self.root
        input_cols = parse_columns(df, input_cols)
        value = df.cols.select(input_cols).to_pandas().values.tolist()

        return value

    @dispatch(str)
    def sort(self, input_col) -> DataFrameType:
        df = self.root
        input_col = parse_columns(df, input_col)
        return df.rows.sort([(input_col, "desc",)])

    @dispatch(str, bool)
    def sort(self, input_col, asc=False) -> DataFrameType:
        df = self.root
        input_col = parse_columns(df, input_col)
        return df.rows.sort([(input_col, "asc" if asc else "desc",)])

    @dispatch(str, str)
    def sort(self, input_col, order="desc") -> DataFrameType:
        df = self.root
        input_col = parse_columns(df, input_col)
        return df.rows.sort([(input_col, order,)])

    @dispatch(dict)
    def sort(self, col_sort) -> DataFrameType:
        df = self.root
        return df.rows.sort([(k, v) for k, v in col_sort.items()])

    @dispatch(list, bool)
    def sort(self, input_col, asc=False) -> DataFrameType:
        df = self.root
        return df.rows.sort(input_col, "asc" if asc else "desc")

    @dispatch(list)
    def sort(self, input_col) -> DataFrameType:
        df = self.root
        return df.rows.sort(input_col)

    @dispatch(list, str)
    def sort(self, input_col, order="desc") -> DataFrameType:
        """
        Sort rows taking into account multiple columns
        :param input_col: column and sort type combination (col_name, "asc")
        :type input_col: list of tuples
        """
        # If a list of columns names are given order this by desc. If you need to specify the order of every
        # column use a list of tuples (col_name, "asc")
        df = self.root
        dfd = df.data
        meta = df.meta

        if is_list_of_str_or_int(input_col):
            t = []
            for col_name in input_col:
                t.append(tuple([col_name, order]))
            input_col = t

        dfd, meta = self._sort_multiple(dfd, meta, input_col)

        return self.root.new(dfd, meta=meta)

    def reverse(self) -> DataFrameType:
        """

        :return:
        """
        dfd = self._reverse(self.root.data)
        return self.root.new(dfd)

    def drop(self, where) -> DataFrameType:
        """
        Drop rows depending on a mask or an expression
        :param where: Mask, expression or name of the column to be taken as mask
        :return: Optimus Dataframe
        """
        df = self.root
        dfd = df.data

        if is_str(where):
            if where in df.cols.names():
                where = df[where]
            else:
                where = eval(where)
        dfd = dfd[where.get_series()==0]
        meta = Meta.action(df.meta, Actions.DROP_ROW.value, df.cols.names())
        return self.root.new(dfd, meta=meta)

    def between_index(self, cols="*", lower_bound=None, upper_bound=None):
        """

        :param columns:
        :param lower_bound:
        :param upper_bound:
        :return:
        """
        dfd = self.root.data
        cols = val_to_list(parse_columns(dfd, cols))
        return self.root.new(dfd[lower_bound: upper_bound][cols])

    def between(self, cols="*", lower_bound=None, upper_bound=None, equal=False, bounds=None, drop=False, how="any") -> DataFrameType:
        
        df = self.root
        
        if bounds is None:
            bounds = [(lower_bound, upper_bound)]

        mask = None

        for bound in bounds:
            if equal:
                _mask = (df[cols] >= bound[0]) & (df[cols] <= bound[1])
            else:
                _mask = (df[cols] > bound[0]) & (df[cols] < bound[1])
            
            if mask is None:
                mask = _mask
            else:
                mask = mask | _mask

        if how=="any":
            mask = mask.mask.any()
        elif how=="all":
            mask = mask.mask.all()
        else:
            RaiseIt.value_error(how, ["any", "all"])

        if drop:
            mask = ~mask
     
        df = df.rows.select(mask)
        
        return df

    def limit(self, count=10) -> DataFrameType:
        """
        Limit the number of rows
        :param count:
        :return:
        """
        return self.root.new(self.root.data[:count])

    @staticmethod
    @abstractmethod
    def unnest(input_cols) -> DataFrameType:
        pass

    def approx_count(self) -> DataFrameType:
        """
        Aprox count
        :return:
        """
        return self.root.rows.count()

    def _mask(self, cols, method, drop=False, how="any", *args, **kwargs) -> DataFrameType:

        df = self.root
        mask = getattr(df.mask, method)(cols=cols, *args, **kwargs)

        if how=="any":
            mask = mask.mask.any()
        elif how=="all":
            mask = mask.mask.all()
        else:
            RaiseIt.value_error(how, ["any", "all"])

        if drop:
            mask = ~mask

        df = df.rows.select(mask)

        return df


    def str(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="str", drop=drop, how=how)

    def int(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="int", drop=drop, how=how)

    def float(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="float", drop=drop, how=how)

    def numeric(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="numeric", drop=drop, how=how)

    def greater_than_equal(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="greater_than_equal", drop=drop, value=value, how=how)

    def greater_than(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="greater_than", drop=drop, value=value, how=how)

    def less_than(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="less_than", drop=drop, value=value, how=how)

    def less_than_equal(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="less_than_equal", drop=drop, value=value, how=how)

    def equal(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="equal", drop=drop, value=value, how=how)

    def not_equal(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="not_equal", drop=drop, value=value, how=how)

    def missing(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="missing", drop=drop, how=how)

    def null(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="null", drop=drop, how=how)

    def none(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="none", drop=drop, how=how)

    def nan(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="nan", drop=drop, how=how)

    def empty(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="empty", drop=drop, how=how)

    def duplicated(self, cols="*", keep="first", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="duplicated", drop=drop, keep=keep, how=how)

    def unique(self, cols="*", keep="first", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="unique", drop=drop, keep=keep, how=how)

    def mismatch(self, cols="*", dtype=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="mismatch", drop=drop, dtype=dtype, how=how)

    def match(self, cols="*", regex=None, dtype=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="match", drop=drop, regex=regex, dtype=dtype, how=how)

    def match_regex(self, cols="*", regex=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="match_regex", drop=drop, regex=regex, how=how)

    def match_dtype(self, cols="*", dtype=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="match_dtype", drop=drop, dtype=dtype, how=how)

    def value_in(self, cols="*", values=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="value_in", drop=drop, values=values, how=how)

    def pattern(self, cols="*", pattern=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="pattern", drop=drop, pattern=pattern, how=how)

    def starts_with(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="starts_with", drop=drop, value=value, how=how)

    def ends_with(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="ends_with", drop=drop, value=value, how=how)

    def contains(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="contains", drop=drop, value=value, how=how)

    def find(self, cols="*", value=None, drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="find", drop=drop, value=value, how=how)

    def email(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="email", drop=drop, how=how)
    
    def ip(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="ip", drop=drop, how=how)
    
    def url(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="url", drop=drop, how=how)
    
    def gender(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="gender", drop=drop, how=how)
    
    def boolean(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="boolean", drop=drop, how=how)
    
    def zip_code(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="zip_code", drop=drop, how=how)
    
    def credit_card_number(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="credit_card_number", drop=drop, how=how)
    
    def datetime(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="datetime", drop=drop, how=how)
    
    def object(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="object", drop=drop, how=how)
    
    def array(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="array", drop=drop, how=how)
    
    def phone_number(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="phone_number", drop=drop, how=how)
    
    def social_security_number(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="social_security_number", drop=drop, how=how)
    
    def http_code(self, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="http_code", drop=drop, how=how)
    
    def expression(self, where=None, cols="*", drop=False, how="any") -> DataFrameType:
        return self._mask(cols, method="expression", drop=drop, how=how, where=where)
    
    # drop functions

    def drop_str(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="str", drop=True, how=how)

    def drop_int(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="int", drop=True, how=how)

    def drop_float(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="float", drop=True, how=how)

    def drop_numeric(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="numeric", drop=True, how=how)
    
    def drop_greater_than_equal(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="greater_than_equal", drop=True, value=value, how=how)

    def drop_greater_than(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="greater_than", drop=True, value=value, how=how)

    def drop_less_than_equal(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="less_than_equal", drop=True, value=value, how=how)

    def drop_less_than(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="less_than", drop=True, value=value, how=how)

    def drop_equal(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="equal", drop=True, value=value, how=how)

    def drop_not_equal(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="not_equal", drop=True, value=value, how=how)

    def drop_missings(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="missing", drop=True, how=how)

    def drop_nulls(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="null", drop=True, how=how)

    def drop_none(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="none", drop=True, how=how)

    def drop_nan(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="nan", drop=True, how=how)

    def drop_empty(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="empty", drop=True, how=how)

    def drop_duplicated(self, cols="*", keep="first", how="any") -> DataFrameType:
        return self._mask(cols, method="duplicated", drop=True, keep=keep, how=how)

    def drop_uniques(self, cols="*", keep="first", how="any") -> DataFrameType:
        """
        Drops first (passed to keep) matches of duplicates and unique values.
        :param cols: 
        :param keep: 
        :param how: 
        :return: Dataframe
        """
        return self._mask(cols, method="unique", drop=True, keep=keep, how=how)

    def drop_mismatch(self, cols="*", dtype=None, how="any") -> DataFrameType:
        return self._mask(cols, method="mismatch", drop=True, dtype=dtype, how=how)

    def drop_match(self, cols="*", regex=None, dtype=None, how="any") -> DataFrameType:
        return self._mask(cols, method="match", drop=True, regex=regex, dtype=dtype, how=how)

    def drop_by_regex(self, cols="*", regex=None, how="any") -> DataFrameType:
        return self._mask(cols, method="match_regex", drop=True, regex=regex, how=how)

    def drop_by_dtype(self, cols="*", dtype=None, how="any") -> DataFrameType:
        return self._mask(cols, method="match_dtype", drop=True, dtype=dtype, how=how)

    def drop_value_in(self, cols="*", values=None, how="any") -> DataFrameType:
        return self._mask(cols, method="value_in", drop=True, values=values, how=how)

    def drop_pattern(self, cols="*", pattern=None, how="any") -> DataFrameType:
        return self._mask(cols, method="pattern", drop=True, pattern=pattern, how=how)

    def drop_starts_with(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="starts_with", drop=True, value=value, how=how)

    def drop_ends_with(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="ends_with", drop=True, value=value, how=how)

    def drop_contains(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="contains", drop=True, value=value, how=how)

    def drop_find(self, cols="*", value=None, how="any") -> DataFrameType:
        return self._mask(cols, method="find", drop=True, value=value, how=how)

    def drop_emails(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="email", drop=True, how=how)
    
    def drop_ips(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="ip", drop=True, how=how)
    
    def drop_urls(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="url", drop=True, how=how)
    
    def drop_genders(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="gender", drop=True, how=how)
    
    def drop_booleans(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="boolean", drop=True, how=how)
    
    def drop_zip_codes(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="zip_code", drop=True, how=how)
    
    def drop_credit_card_numbers(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="credit_card_number", drop=True, how=how)
    
    def drop_datetimes(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="datetime", drop=True, how=how)
    
    def drop_objects(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="object", drop=True, how=how)
    
    def drop_arrays(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="array", drop=True, how=how)
    
    def drop_phone_numbers(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="phone_number", drop=True, how=how)
    
    def drop_social_security_numbers(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="social_security_number", drop=True, how=how)
    
    def drop_http_codes(self, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="http_code", drop=True, how=how)

    def drop_by_expression(self, where=None, cols="*", how="any") -> DataFrameType:
        return self._mask(cols, method="expression", drop=True, how=how, where=where)


# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.1.1
#   kernel_info:
#     name: python3
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # This notebook create the tests in python code. All this cells must be run to executed the tests

# %load_ext autoreload
# %autoreload 2

# + {"outputHidden": false, "inputHidden": false}
import sys
sys.path.append("../..")
# -

from optimus import Optimus
from optimus.helpers.test import Test

op = Optimus(master='local', verbose=True)

# +
import pandas as pd
from pyspark.sql.types import *
from datetime import date, datetime


cols = [
        ("names", "str"),
        ("height(ft)", ShortType()),
        ("function", "str"),
        ("rank", ByteType()),
        ("age", "int"),
        ("weight(t)", "float"),
        "japanese name",
        "last position seen",
        "date arrival",
        "last date seen",
        ("attributes", ArrayType(FloatType())),
        ("Date Type", DateType()),
        ("timestamp", TimestampType()),
        ("Cybertronian", BooleanType()),
        ("function(binary)", BinaryType()),
        ("NullType", NullType())

    ]

rows = [
        ("Optimus", -28, "Leader", 10, 5000000, 4.30, ["Inochi", "Convoy"], "19.442735,-99.201111", "1980/04/10",
         "2016/09/10", [8.5344, 4300.0], date(2016, 9, 10), datetime(2014, 6, 24), True, bytearray("Leader", "utf-8"),
         None),
        ("bumbl#ebéé  ", 17, "Espionage", 7, 5000000, 2.0, ["Bumble", "Goldback"], "10.642707,-71.612534", "1980/04/10",
         "2015/08/10", [5.334, 2000.0], date(2015, 8, 10), datetime(2014, 6, 24), True, bytearray("Espionage", "utf-8"),
         None),
        ("ironhide&", 26, "Security", 7, 5000000, 4.0, ["Roadbuster"], "37.789563,-122.400356", "1980/04/10",
         "2014/07/10", [7.9248, 4000.0], date(2014, 6, 24), datetime(2014, 6, 24), True, bytearray("Security", "utf-8"),
         None),
        ("Jazz", 13, "First Lieutenant", 8, 5000000, 1.80, ["Meister"], "33.670666,-117.841553", "1980/04/10",
         "2013/06/10", [3.9624, 1800.0], date(2013, 6, 24), datetime(2014, 6, 24), True,
         bytearray("First Lieutenant", "utf-8"), None),
        ("Megatron", None, "None", 10, 5000000, 5.70, ["Megatron"], None, "1980/04/10", "2012/05/10", [None, 5700.0],
         date(2012, 5, 10), datetime(2014, 6, 24), True, bytearray("None", "utf-8"), None),
        ("Metroplex_)^$", 300, "Battle Station", 8, 5000000, None, ["Metroflex"], None, "1980/04/10", "2011/04/10",
         [91.44, None], date(2011, 4, 10), datetime(2014, 6, 24), True, bytearray("Battle Station", "utf-8"), None),
        

    ]
source_df = op.create.df(cols ,rows)
source_df.display()
# -




# ### End Init Section

# ## Ext

from pyspark.ml.linalg import Vectors

t = Test(op, source_df, "df_ext", imports=["from pyspark.ml.linalg import Vectors, VectorUDT, DenseVector",
                                        "import numpy as np",
                                        "nan = np.nan",
                                        "import datetime",
                                        "from pyspark.sql import functions as F",
                                        "from optimus.profiler.profiler import Profiler",
                                        "null = None",
                                        "true = True"], path = "df_ext", final_path="..")

# +
from pyspark.sql import functions as F


def func(col_name, attrs):
    return F.col(col_name) * 2

numeric_col = "height(ft)"
numeric_col_B = "rank"
numeric_col_C = "rank"
string_col = "function"
date_col = "date arrival"
date_col_B = "last date seen"
new_col = "new col"
array_col = "attributes"
# -

source_df.rows.count()

t.create(None, "ext.sample", None, 'df', None, 5, True)
# t.create(source_df_string_to_index, "cols.string_to_index", None, "df", None, "rank")

t.run()

t.create(None, "ext.stratified_sample", None, 'df', None, numeric_col_B, 12)

t.run()





import glob
import ntpath
import uuid
import zipfile
from pathlib import Path
from zipfile import ZipFile

import pandas as pd
import pandavro as pdx
import psutil

from optimus.engines.base.io.load import BaseLoad
from optimus.engines.base.meta import Meta
from optimus.engines.pandas.dataframe import PandasDataFrame
from optimus.helpers.functions import prepare_path, unquote_path
from optimus.helpers.logger import logger
from optimus.infer import is_str, is_list, is_url


class Load(BaseLoad):

    def __init__(self, op):
        self.op = op

    @staticmethod
    def json(path, multiline=False, n_rows=False, *args, **kwargs):
        """
        Loads a dataframe from a json file.
        :param path: path or location of the file.
        :param multiline:

        :return:
        """
        if n_rows:
            kwargs["n_rows"] = n_rows
            kwargs["lines"] = True

        if is_str(path):
            try:
                path = unquote_path(path)
                local_file_names = prepare_path(path, "json")
                df_list = []

                for file_name, j in local_file_names:
                    df = pd.read_json(file_name, lines=multiline, *args, **kwargs)
                    df_list.append(df)
                df = pd.concat(df_list, axis=0, ignore_index=True)
                df = PandasDataFrame(df)
                df.meta = Meta.set(df.meta, "file_name", local_file_names[0])
            except IOError as error:
                logger.print(error)
                raise

        else:
            df = pd.read_json(path, lines=multiline, *args, **kwargs)
            df = PandasDataFrame(df)

        return df


    @staticmethod
    def tsv(path, header=True, infer_schema=True, *args, **kwargs):
        """
        Loads a dataframe from a tsv file.
        :param path: path or location of the file.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        It requires one extra pass over the data. True default.

        :return:
        """

        return Load.csv(path, sep='\t', header=header, infer_schema=infer_schema, *args, **kwargs)


    @staticmethod
    def csv(filepath_or_buffer, sep=",", header=True, infer_schema=True, encoding="UTF-8", n_rows=None,
            null_value="None", quoting=3, lineterminator='\r\n', error_bad_lines=False, cache=False, na_filter=False,
            storage_options=None, conn=None,
            *args, **kwargs):
        """
        Loads a dataframe from a csv file. It is the same read.csv Spark function with some predefined
        params


        :param encoding:
        :param storage_options:
        :param quoting:
        :param filepath_or_buffer: path or location of the file.
        :param sep: usually delimiter mark are ',' or ';'.
        :param header: tell the function whether dataset has a header row. True default.
        :param infer_schema: infers the input schema automatically from data.
        :param n_rows:
        :param null_value:
        :param cache:
        :param na_filter:
        :param lineterminator:
        :param error_bad_lines:
        :param conn:
        It requires one extra pass over the data. True default.

        :return dataFrame
        """
        if not is_url(filepath_or_buffer):
            filepath_or_buffer = glob.glob(unquote_path(filepath_or_buffer))
            meta = {"file_name": filepath_or_buffer, "name": ntpath.basename(filepath_or_buffer[0])}
        else:
            meta = {"file_name": filepath_or_buffer, "name": ntpath.basename(filepath_or_buffer)}

        try:

            # Pandas do not support \r\n terminator.
            if lineterminator and lineterminator.encode(encoding='UTF-8', errors='strict') == b'\r\n':
                lineterminator = None

            if conn is not None:
                filepath_or_buffer = conn.path(filepath_or_buffer)
                storage_options = conn.storage_options
            else:
                storage_options = None

            if kwargs.get("chunk_size") == "auto":
                ## Chunk size is going to be 75% of the memory available
                kwargs.pop("chunk_size")
                kwargs["chunksize"] = psutil.virtual_memory().free * 0.75

            def _read(_filepath_or_buffer):
                return pd.read_csv(_filepath_or_buffer, sep=sep, header=0 if header else -1, encoding=encoding,
                                   nrows=n_rows,
                                   quoting=quoting, lineterminator=lineterminator, error_bad_lines=error_bad_lines,
                                   na_filter=na_filter, index_col=False, storage_options=storage_options, *args,
                                   **kwargs)

            if is_list(filepath_or_buffer):
                df = pd.DataFrame()
                for f in filepath_or_buffer:
                    df = df.append(_read(f))
            else:
                df = _read(filepath_or_buffer)

            if isinstance(df, pd.io.parsers.TextFileReader):
                df = df.get_chunk()

            df = PandasDataFrame(df)

            df.meta = Meta.set(df.meta, value=meta)

        except IOError as error:
            print(error)
            logger.print(error)
            raise

        return df


    @staticmethod
    def parquet(path, columns=None, storage_options=None, conn=None, *args, **kwargs):
        """
        Loads a dataframe from a parquet file.
        :param path: path or location of the file. Must be string dataType
        :param columns: select the columns that will be loaded. In this way you do not need to load all the dataframe
        :param args: custom argument to be passed to the spark parquet function
        :param kwargs: custom keyword arguments to be passed to the spark parquet function
        """

        path = unquote_path(path)

        # file, file_name = prepare_path(path, "parquet")[0]

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        try:
            df = pd.read_parquet(path, columns=columns, engine='pyarrow', storage_options=storage_options, **kwargs)
            df = PandasDataFrame(df)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path)})

        except IOError as error:
            logger.print(error)
            raise

        return df


    @staticmethod
    def avro(path, storage_options=None, conn=None, *args, **kwargs):
        """
        Loads a dataframe from a avro file.
        :param storage_options:
        :param path: path or location of the file. Must be string dataType
        :param args: custom argument to be passed to the spark avro function
        :param kwargs: custom keyword arguments to be passed to the spark avro function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "avro")[0]

        try:
            df = pdx.read_avro(file_name, storage_options=storage_options, *args, **kwargs)
            df = PandasDataFrame(df)
            df.meta = Meta.set(df.meta, value={"file_name": path, "name": ntpath.basename(path)})

        except IOError as error:
            logger.print(error)
            raise

        return df


    @staticmethod
    def excel(path, sheet_name=0, storage_options=None, conn=None, *args, **kwargs):
        """
        Loads a dataframe from a excel file.
        :param path: Path or location of the file. Must be string dataType
        :param sheet_name: excel sheet name
        :param args: custom argument to be passed to the excel function
        :param kwargs: custom keyword arguments to be passed to the excel function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "xls")[0]

        try:
            df = pd.read_excel(file, sheet_name=sheet_name, storage_options=storage_options, *args, **kwargs)

            # Parse object column data type to string to ensure that Spark can handle it. With this we try to reduce
            # exception when Spark try to infer the column data type
            col_names = list(df.select_dtypes(include=['object']))

            column_dtype = {}
            for col in col_names:
                column_dtype[col] = str

            # Convert object columns to string
            df = df.astype(column_dtype)

            # Create spark data frame
            df = PandasDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", ntpath.basename(file_name))
        except IOError as error:
            logger.print(error)
            raise

        return df


    @staticmethod
    def orc(path, columns, storage_options=None, conn=None, *args, **kwargs):
        """
        Loads a dataframe from a avro file.
        :param path: path or location of the file. Must be string dataType
        :param args: custom argument to be passed to the spark avro function
        :param kwargs: custom keyword arguments to be passed to the spark avro function
        """

        path = unquote_path(path)

        if conn is not None:
            path = conn.path(path)
            storage_options = conn.storage_options

        file, file_name = prepare_path(path, "orc")[0]

        try:
            df = pdx.read_orc(file_name, columns, storage_options=storage_options)
            df = PandasDataFrame(df)
            df.meta = Meta.set(df.meta, "file_name", file_name)

        except IOError as error:
            logger.print(error)
            raise

        return df


    @staticmethod
    def zip(zip_path, filename, dest=None, merge=False, storage_options=None, conn=None, *args, **kwargs):
        if dest is None:
            dest = str(uuid.uuid4()) + "/"

        zip_path = glob.glob(zip_path)

        dest = Path(dest).expanduser()

        # if csv concat all files
        # if json multilie concat files

        for filename in zip_path:
            # print(filename)
            with zipfile.ZipFile(filename) as zf:
                zf.infolist()
                for member in zf.infolist():
                    # print(member.filename)
                    try:
                        zf.extract(member, dest)
                    except zipfile.error as e:
                        pass



import re

from ordered_set import OrderedSet

# from optimus.helpers.check import is_spark_dataframe, is_pandas_dataframe, is_dask_dataframe, is_cudf_dataframe

from optimus.helpers.core import val_to_list, one_list_to_val
from optimus.helpers.logger import logger
from optimus.helpers.parser import parse_dtypes
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_list_value, is_tuple, is_list_of_str, is_list_of_list, is_list_of_tuples, is_str, is_list, \
    is_int, is_list_of_int


def replace_columns_special_characters(df, replace_by="_"):
    """
    Remove special character from Spark column name
    :param df: Spark Dataframe
    :param replace_by: character or string that will replace the matched character
    :return:
    """

    def replace_multiple_characters(string, to_be_replaced, replace_by):
        """
        Replace multiple single characters in s string
        :param string:
        :param to_be_replaced: Character to be replaced
        :param replace_by: character or string that will replace the matched character
        :return:
        """
        # Iterate over the strings to be replaced
        for elem in to_be_replaced:
            # Check if string is in the main string
            if elem in string:
                # Replace the string
                string = string.replace(elem, replace_by)
        return string

    for col_name in df.cols.names():
        df = df.cols.rename(col_name, replace_multiple_characters(col_name, ["."], replace_by))
    return df


def escape_columns(columns):
    """
    Add a backtick to a columns name to prevent the dot in name problem
    :param columns:
    :return:
    """

    escaped_columns = []
    if is_list_value(columns):
        for col in columns:
            # Check if the column is already escaped
            if col[0] != "`" and col[len(col) - 1] != "`":
                escaped_columns.append("`" + col + "`")
            else:
                escaped_columns.append(col)
    else:
        # Check if the column is already escaped
        if columns[0] != "`" and columns[len(columns) - 1] != "`":
            escaped_columns = "`" + columns + "`"
        else:
            escaped_columns.append(columns)

    return escaped_columns


def get_output_cols(cols, output_cols=None, merge=False, auto_increment=False):
    """
    Construct output columns taking the input columns.
    If it receive a list of input columns and on output column the function will append the output_col name to the input cols list

    If
    cols = ['col1'] to output_cols = None
    cols = ['col1'] to output_cols = ['col1']


    If merge is True

    For 1 column
    cols = ['col1'] to output_cols = 'new'
    Result:
    cols = ['col1'] to output_cols = ['col1_new']

    For multiple columns
    cols = ['col1', 'col2'...'col3'] to output_cols = 'new'
    Result:
    cols = ['col1', 'col2',...'col3'] to output_cols = ['col1_new', 'col2_new',...'col3_new']

    else merge is False
    For 1 column
    cols = ['col1'] to output_cols = 'new'
    cols = ['col1'] to output_cols = ['new']

    :param cols:
    :param output_cols:
    :param merge:
    :param auto_increment:
    :return:
    """

    # if is_list(cols) and is_list(output_cols):
    #     if len(cols) != len(output_cols):
    #         RaiseIt.length_error(cols, output_cols)

    if output_cols is None:
        output_cols = val_to_list(cols)
    elif callable(output_cols):
        output_cols = list(map(output_cols, cols))
    else:
        output_cols = val_to_list(output_cols)

        # if auto_increment is not None:

    if auto_increment is True:
        # cols = cols * auto_increment
        # output_cols = val_to_list(output_cols)
        # print("LEAN @", len(cols)/auto_increment)
        # r = int(len(cols) / auto_increment)
        r = list(range(auto_increment)) * 2
        output_cols = [col_name + "_" + str(i) for i, col_name in zip(r, cols)]
    elif merge is True:
        output_cols = val_to_list(output_cols)
        output_cols = list([name_col(input_col, output_cols) for input_col in cols])

    return output_cols


def parse_columns(df, cols_args, is_regex=None, filter_by_column_dtypes=None, accepts_missing_cols=False, invert=False):
    """
    Return a list of columns and check that columns exists in the spark
    Accept '*' as parameter in which case return a list of all columns in the spark.
    Also accept a regex.
    If a list of tuples return to list. The first element is the columns name the others element are params.
    This params can be used to create custom transformation functions. You can find and example in cols().cast()
    :param df: Dataframe in which the columns are going to be checked
    :param cols_args: Accepts * as param to return all the string columns in the spark
    :param is_regex: Use True is col_attrs is a regex
    :param filter_by_column_dtypes: A data type for which a columns list is going be filtered
    :param accepts_missing_cols: if true not check if column exist in the spark
    :param invert: Invert the final selection. For example if you want to select not integers

    :return: A list of columns string names
    """
    
    # if columns value is * get all dataframes columns

    df_columns = df.cols._names()
    if is_regex is True:
        r = re.compile(cols_args)
        cols = list(filter(r.match, df_columns))

    elif cols_args == "*" or cols_args is None:
        cols = df_columns
    elif is_int(cols_args):
        cols = val_to_list(df_columns[cols_args])
    elif is_list_of_int(cols_args):
        cols = list(df_columns[i] for i in cols_args)

    elif is_tuple(cols_args) or is_list_of_tuples(cols_args):
        # In case we have a list of tuples we use the first element of the tuple is taken as the column name
        # and the rest as params. We can use the param in a custom function as follow
        # def func(attrs): attrs return (1,2) and (3,4)
        #   return attrs[0] + 1
        # df.cols().apply([('col_1',1,2),('cols_2', 3 ,4)], func)

        # Verify if we have a list with tuples

        cols_args = val_to_list(cols_args)
        # Extract a specific position in the tuple
        cols = [(i[0:1][0]) for i in cols_args]
        attrs = [(i[1:]) for i in cols_args]
    else:
        # if not a list convert to list
        cols = val_to_list(cols_args)
        # Get col name from index
        cols = [c if is_str(c) else df_columns[c] for c in cols]

    # Check for missing columns
    if accepts_missing_cols is False:
        check_for_missing_columns(df, cols)

    # Filter by column data type
    filter_by_column_dtypes = val_to_list(filter_by_column_dtypes)
    if is_list_of_list(filter_by_column_dtypes):
        filter_by_column_dtypes = [item for sublist in filter_by_column_dtypes for item in sublist]

    columns_residual = None

    # If necessary filter the columns by data type
    if filter_by_column_dtypes:
        # Get columns for every data type
        columns_filtered = filter_col_name_by_dtypes(df, filter_by_column_dtypes)

        # Intersect the columns filtered per data type from the whole spark with the columns passed to the function
        final_columns = list(OrderedSet(cols).intersection(columns_filtered))

        # This columns match filtered data type
        columns_residual = list(OrderedSet(cols) - OrderedSet(columns_filtered))

    else:
        final_columns = cols

    cols_params = []
    if invert:
        final_columns = list(OrderedSet(df_columns) - OrderedSet(final_columns))

    cols_params = final_columns

    if columns_residual:
        logger.print("%s %s %s", ",".join(escape_columns(columns_residual)),
                     "column(s) was not processed because is/are not",
                     ",".join(filter_by_column_dtypes))

    # if because of filtering we got 0 columns return None
    if len(cols_params) == 0:
        cols_params = None
        logger.print("Outputting 0 columns after filtering. Is this expected?")

    return cols_params


def prepare_columns(df, cols: [str, list], output_cols: [str, list] = None, is_regex=None,
                    filter_by_column_dtypes=None, accepts_missing_cols=False, invert: bool = False, default=None,
                    cols_dict=None, auto_increment=False, args=None):
    """
    One input columns- > Same output column. lower(), upper()
    One input column -> One output column. copy()
    One input column -> Multiple output column. unnest()
    Multiple input columns -> One output column. nest()
    Multiple input columns -> Multiple output columns. lower(), upper()

    For multiple output we can pass a string, a list o columns or just enumerate de output column.
    See get_output_cols() for more info

    Accepts Return an iterator with input and output columns
    :param df: dataframe against to check that the input columns are valid
    :param cols: intput columns names
    :param output_cols: output columns names
    :param is_regex: input columns is a regex
    :param filter_by_column_dtypes: filter column selection by data type
    :param accepts_missing_cols: dont check the input columns exist
    :param invert: Invert selection
    :param default: Default column name if output_cols is not provider
    :param cols_dict: In case you have a input output dictionary already defined. {incol:outcol1 , incol2:outcol2}
    :param auto_increment:
    :param merge:
    :return:
    """
    if cols_dict:
        result = zip(*cols_dict)
    else:
        cols = parse_columns(df, cols, is_regex, filter_by_column_dtypes,
                                   accepts_missing_cols, invert)
        merge = False
        if output_cols is None and default is not None:
            output_cols = default
            merge = True

        elif auto_increment is not False:
            cols = cols * auto_increment

        if output_cols is not None and (len(cols) != len(val_to_list(output_cols))):
            merge = True

        output_cols = get_output_cols(cols, output_cols, merge=merge, auto_increment=auto_increment)

        if args is None:
            result = zip(cols, output_cols)
        else:
            args = val_to_list(args)
            if len(args) == 1:
                args = args * len(cols)

            result = zip(cols, output_cols, args)
    return result


def check_column_numbers(columns, number=0):
    """
    Check if the columns number match number expected
    :param columns:
    :param number: Number of columns to check
    :return:
    """
    if columns is None:
        RaiseIt.value_error(columns, ["str", "list"],
                            extra_text="Maybe the columns selected do not match a specified datatype filter.")

    if isinstance(columns, zip):
        columns = list(columns)

    count = list(columns)

    if number == "*":
        if not len(columns) >= 1:
            RaiseIt.value_error(len(columns), ["1 or greater"])
    elif number == ">1":
        if not len(columns) > 1:
            RaiseIt.value_error(len(columns), ["more than 1"])
    elif len(columns) != number:
        RaiseIt.value_error(count, "{} columns, {} needed".format(number, columns))
    # elif isinstance(columns,zip):


def validate_columns_names(df, col_names: [str, list], index=0):
    """
    Check if a string or list of string are valid spark columns
    :param df: Dataframe to be analyzed
    :param col_names: columns names to be checked
    :param index:
    :return:
    """

    columns = val_to_list(col_names)

    if is_list_of_tuples(columns):
        columns = [c[index] for c in columns]

    # Remove duplicates in the list

    if is_list_of_str(columns):
        columns = OrderedSet(columns)

    check_for_missing_columns(df, columns)

    return True


def check_for_missing_columns(df, col_names):
    """
    Check if the columns you want to select exits in the dataframe
    :param df: Dataframe to be checked
    :param col_names: cols names to
    :return:
    """
    _col_names = df.cols._names()
    missing_columns = list(OrderedSet(col_names) - OrderedSet(_col_names))

    if len(missing_columns) > 0:
        RaiseIt.value_error(missing_columns, _col_names)
    return False


def filter_col_name_by_dtypes(df, data_type):
    """
    Return column names filtered by the column data type
    :param df: Dataframe which columns are going to be filtered
    :param data_type: Datatype used to filter the column.
    :type data_type: str or list
    :return:
    """
    data_type = parse_dtypes(df, data_type)
    data_type = val_to_list(data_type)
    # Filter columns by data type
    result = []
    for col_name in df.cols.names():
        found_data_type = df.cols.schema_dtype(col_name)        
        if any([dt in found_data_type for dt in data_type]):
            result.append(col_name)
    return result


def name_col(col_names: str, append: str = None) -> str:
    """
    Whenever you want to name and output use this function. This ensure that we use and Standard when naming
    :param col_names: Column name
    :param append: string to be appended
    :return:
    """
    separator = "_"

    col_names = val_to_list(col_names)
    if len(col_names) > 1 and append is None:
        output_col = ('_'.join(str(elem) for elem in col_names))
    elif len(col_names) > 1:
        output_col = ('_'.join(str(elem) for elem in col_names))[:10] + separator
    else:
        output_col = one_list_to_val(col_names)

    if append is not None:
        append = separator + str(one_list_to_val(append))
    else:
        append = ""
    return output_col + append



from nltk.util import ngrams
from pyspark.sql import functions as F

from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.json import dump_json
from optimus.engines.base.ml.contants import CLUSTER_COL, COUNT_COL, RECOMMENDED_COL, CLUSTER_COUNT_COL, \
    FINGERPRINT_COL, \
    CLUSTER_SUM_COL


def fingerprint(df, input_cols):
    """
    Create the fingerprint for a column
    :param df: Dataframe to be processed
    :param input_cols: Column to be processed
    :return:
    """

    # https://github.com/OpenRefine/OpenRefine/blob/master/main/src/com/google/refine/clustering/binning/FingerprintKeyer.java#L56
    def _split_sort_remove_join(value):
        """
        Helper function to split, remove duplicates, sort and join back together
        """
        # Split into whitespace-separated token
        # print("value", type(value), value)
        split_key = value.split()

        # Sort and remove duplicated items
        split_key = sorted(set(split_key))

        # join the tokens back together
        return "".join(split_key)

    input_cols = parse_columns(df, input_cols)
    for input_col in input_cols:
        output_col = name_col(input_col, FINGERPRINT_COL)
        df = (df
              .cols.trim(input_col, output_col)
              .cols.lower(output_col)
              .cols.remove_special_chars(output_col)
              .cols.normalize_chars(output_col)
              .cols.apply(output_col, _split_sort_remove_join, "string", mode="map")
              )
    return df


def n_gram_fingerprint(df, input_cols, n_size=2):
    """
    Calculate the ngram for a fingerprinted string
    :param df: Dataframe to be processed
    :param input_cols: Columns to be processed
    :param n_size:
    :return:
    """

    def calculate_ngrams(value, args):
        # remove white spaces
        ngram = list(ngrams(value, n_size))

        # sort and remove duplicated
        ngram = sorted(set(ngram))

        _result = ""
        for item in ngram:
            for i in item:
                _result = _result + i

        # join the tokens back together
        _result = "".join(_result)

        return _result

    input_cols = parse_columns(df, input_cols)

    for input_col in input_cols:
        ngram_fingerprint_col = name_col(input_col, FINGERPRINT_COL)
        # ngram_fingerprint_col = name_col(input_col, NGRAM_FINGERPRINT_COL)

        df = (df
              .cols.copy(input_col, ngram_fingerprint_col)
              .cols.lower(ngram_fingerprint_col)
              .cols.remove_white_spaces(ngram_fingerprint_col)
              .cols.apply(ngram_fingerprint_col, calculate_ngrams, "string", output_cols=ngram_fingerprint_col)
              .cols.remove_special_chars(ngram_fingerprint_col)
              .cols.normalize_chars(ngram_fingerprint_col)

              )

    return df


def fingerprint_cluster(df, input_cols, output: str = "dict"):
    return base_clustering_function(df, input_cols, output, func=fingerprint, args=[input_cols])


def n_gram_fingerprint_cluster(df, input_cols, n_size=2, output: str = "dict"):
    return base_clustering_function(df, input_cols, output, func=n_gram_fingerprint, args=[input_cols, n_size])


def base_clustering_function(df, input_cols, output, func=None, args=None):
    """
    Cluster a dataframe column based on the Fingerprint algorithm
    :return:
    """

    # df = self.df
    input_cols = parse_columns(df, input_cols)
    result = {}
    for input_col in input_cols:
        output_col = name_col(input_col, FINGERPRINT_COL)
        count_col = name_col(input_col, COUNT_COL)
        # Instead of apply the fingerprint to the whole data set we group by names
        df = (df
              .groupBy(input_col)
              .agg(F.count(input_col).alias(count_col))
              .select(count_col, input_col)
              ).h_repartition(1)

        # Calculate the fingerprint
        recommended_col = name_col(input_col, RECOMMENDED_COL)
        cluster_col = name_col(input_col, CLUSTER_COL)
        cluster_sum_col = name_col(input_col, CLUSTER_SUM_COL)
        cluster_count_col = name_col(input_col, CLUSTER_COUNT_COL)

        # Apply function cluster
        df = func(df, *args)

        df = df.withColumn(cluster_col, F.create_map([input_col, count_col]))

        df = df.groupBy(output_col).agg(F.max(F.struct(F.col(count_col), F.col(input_col))).alias(recommended_col),
                                        F.collect_list(cluster_col).alias(cluster_col),
                                        F.count(output_col).alias(cluster_count_col),
                                        F.sum(count_col).alias(cluster_sum_col))
        df = df.select(recommended_col + "." + input_col, cluster_col, cluster_count_col, cluster_sum_col)

        for row in df.collect():
            _row = list(row.asDict().values())
            # List of dict to dict
            flatted_dict = {k: v for element in _row[1] for k, v in element.items()}
            result[_row[0]] = {"similar": flatted_dict, "count": _row[2], "sum": _row[3]}

    if output == "json":
        result = dump_json(result)
    return result



"""
Helpers to check if an object match a date type
"""
import re

import pandas as pd

from optimus.helpers.core import val_to_list, one_list_to_val
# TODO: can be confused with is_type
from optimus.helpers.parser import parse_dtypes
from optimus.helpers.raiseit import RaiseIt
# from optimus.new_optimus import SparkDataFrame


def has_(value, _type):
    """
    Check if a list has a element of a specific data type
    :param value: list
    :param _type: data type to check
    :return:
    """
    return any(isinstance(elem, _type) for elem in value)


def is_column_a(df, column=None, dtypes="str"):
    """
    Check if column match a list of data types
    :param df: spark or dask dataframe
    :param column: column to be compared with
    :param dtypes: types to be checked
    :return:
    """
    column = val_to_list(column)

    if len(column) > 1:
        RaiseIt.length_error(column, 1)
    data_type = tuple(val_to_list(parse_dtypes(df, dtypes)))
    column = one_list_to_val(column)

    # Filter columns by data type
    # print("df",type(df),df)
    v = df.cols.schema_dtype(column)

    if is_spark_dataframe(df.data):
        result = isinstance(v, data_type)
    elif is_dask_dataframe(df):
        result = v in data_type
    else:
        result = None
    return result


#
# def is_column_a(df, column, dtypes):
#     """
#     Check if column match a list of data types
#     :param df: spark
#     :param column: column to be compared with
#     :param dtypes: types to be checked
#     :return:
#     """
#     column = val_to_list(column)
#
#     if len(column) > 1:
#         RaiseIt.length_error(column, 1)
#
#     data_type = tuple(val_to_list(parse_spark_dtypes(df, dtypes)))
#     column = one_list_to_val(column)
#
#     # Filter columns by data type
#     return isinstance(df.schema[column].dataType, data_type)
# def is_cudf_series(value):
#     return cudf.core.series.Series

# def is_cudf_dataframe(value):
#     from cudf.core import DataFrame as CUDFDataFrame
#     return isinstance(value, CUDFDataFrame)
#
#
# def is_cudf_series(value):
#     import cudf
#     return isinstance(value, cudf.core.series.Series)
#
#
def is_dask_cudf_dataframe(value):
    """
    Check if an object is a Dask cuDF DataFrame
    :param value:
    :return:
    """
    try:
        from dask_cudf.core import DataFrame as DaskCUDFDataFrame
        return isinstance(value, DaskCUDFDataFrame)
    except:
        return False

#
# def is_dask_cudf_series(value):
#     from dask_cudf.core import Series as DaskCUDFSeries
#     return isinstance(value, DaskCUDFSeries)


def is_dask_dataframe(value):
    """
    Check if an object is a Dask DataFrame
    :param value:
    :return:
    """
    from dask.dataframe.core import DataFrame as DaskDataFrame
    return isinstance(value, DaskDataFrame)


def is_dask_series(value):
    """
    Check if an object is a Dask DataFrame
    :param value:
    :return:
    """
    from dask.dataframe.core import Series as DaskSeries
    return isinstance(value, DaskSeries)


def is_spark_dataframe(value):
    """
    Check if an object is a Spark DataFrame
    :param value:
    :return:
    """

    from pyspark.sql import DataFrame as SparkDataFrame
    return isinstance(value, SparkDataFrame)


def is_pandas_dataframe(value):
    """
    Check if an object is a Pandas DataFrame
    :param value:
    :return:
    """
    return isinstance(value, pd.DataFrame)


def is_pandas_series(value):
    """
    Check if an object is a Pandas DataFrame
    :param value:
    :return:
    """
    return isinstance(value, pd.Series)




def equal_function(f1, f2):
    f2 = val_to_list(f2)
    for func in f2:
        if f1.__name__ == func.__name__:
            return True
    return False


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell in ['ZMQInteractiveShell', 'Shell']:
            return True
        else:
            return False
    except NameError:
        return False


import sys
sys.path.append('../optimus')

from optimus.helpers.core import val_to_list

alias_map_list = [
  
  # engine methods
  
  [["optimus", "engine", "createEngine"], "Optimus"],

  # op methods

  [["createDataframe", "dataframe"], "create.dataframe"],

  # connection methods

  ["connectMYSQL", "connect.mysql"],
  ["connectPostgres", "connect.postgres"],
  ["connectMSSQL", "connect.mssql"],
  ["connectRedshift", "connect.redshift"],
  ["connectSQLite", "connect.sqlite"],
  ["connectBigQuery", "connect.bigquery"],
  ["connectPresto", "connect.presto"],
  ["connectCassandra", "connect.cassandra"],
  ["connectRedis", "connect.redis"],
  ["connectOracle", "connect.oracle"],
  ["connectS3", "connect.s3"],
  ["connectLocal", "connect.local"],
  ["connectHDFS", "connect.hdfs"],
  ["connectGCS", "connect.gcs"],
  ["connectMAS", "connect.mas"],

  # string clustering methods

  ["setSuggestion", "set_suggestion"],
  ["setSuggestions", "set_suggestions"],

  # save methods

  ["saveFile", "save.file"],
  ["saveCSV", "save.csv" ],
  ["saveXML", "save.xml" ],
  ["saveJSON", "save.json" ],
  ["saveExcel", "save.excel" ],
  ["saveAvro", "save.avro" ],
  ["saveParquet", "save.parquet" ],
  ["saveORC", "save.orc" ],
  ["saveHDF5", "save.hdf5" ],

  # load methods

  ["loadFile", "load.file"],
  ["loadCSV", { "operation": "load.file", "file_type": "csv" }],
  ["loadXML", { "operation": "load.file", "file_type": "xml" }],
  ["loadJSON", { "operation": "load.file", "file_type": "json" }],
  ["loadExcel", { "operation": "load.file", "file_type": "excel" }],
  ["loadAvro", { "operation": "load.file", "file_type": "avro" }],
  ["loadParquet", { "operation": "load.file", "file_type": "parquet" }],
  ["loadORC", { "operation": "load.file", "file_type": "orc" }],
  ["loadZIP", { "operation": "load.file", "file_type": "zip" }],
  ["loadHDF5", { "operation": "load.file", "file_type": "hdf5" }],

  # methods

  ["profile", "profile"],
  [["profileColumns", "profile.columns"], "profile.columns"],
  [["profileStats", "profile.stats"], "profile.stats"],
  [["display", "toDict", "toObject"], "to_dict"],
  ["toJSON", "to_json"],
  ["size", "cols.size"],
  ["optimize", "cols.optimize"],
  ["run", "cols.run"],
  ["query", "cols.query"],
  ["partitions", "cols.partitions"],
  ["repartition", "cols.repartition"],
  [["print", "ascii"], "ascii"],
  ["reset", "cols.reset"],
  [["agg", "aggregate", "groupyby", "groupBy"], "agg"],
  ["report", "cols.report"],
  [["selectColumns", "select", "cols.select"], "cols.select"],
  [["copyColumns", "copy", "cols.copy"], "cols.copy"],
  [["dropColumns", "drop", "cols.drop"], "cols.drop"],
  [["keepColumns", "keep", "cols.keep"], "cols.keep"],
  [["renameColumns", "cols.rename"], "cols.rename"],
  [["names", "columnNames", "columns"], "cols.names"],
  [["datatypes", "dtypes", "cols.dtypes"], "cols.dtypes"],
  [["patterCounts", "pattern_counts"], "cols.pattern_counts"],
  ["mad", "cols.mad"],
  ["min", "cols.min"],
  ["max", "cols.max"],
  ["mode", "cols.mode"],
  ["range", "cols.range"],
  ["percentile", "cols.percentile"],
  ["median", "cols.median"],
  ["kurtosis", "cols.kurtosis"],
  ["skew", "cols.skew"],
  ["mean", "cols.mean"],
  ["sum", "cols.sum"],
  ["cumsum", "cols.cumsum"],
  ["cumprod", "cols.cumprod"],
  ["cummax", "cols.cummax"],
  ["cummin", "cols.cummin"],
  ["var", "cols.var"],
  ["std", "cols.std"],
  ["iqr", "cols.iqr"],
  [["countNA", "count_na"], "cols.count_na"],
  ["unique", "cols.unique"],
  [["countUniques", "count_uniques"], "cols.count_uniques"],
  ["heatmap", "cols.heatmap"],
  ["hist", "cols.hist"],
  [["countMismatch", "count_mismatch"], "cols.count_mismatch"],
  [["countByDtypes", "count_by_dtypes"], "cols.count_by_dtypes"],
  ["quality", "cols.quality"],
  [
    ["inferProfilerDtypes", "infer_profiler_dtypes"],
    "cols.infer_profiler_dtypes",
  ],
  ["frequency", "cols.frequency"],
  ["boxplot", "cols.boxplot"],
  [["countZeros", "count_zeros"], "cols.count_zeros"],
  [["countRows", "length", "len"], "rows.count"],
  [["approxCount", "approx_count"], "rows.approx_count"],
  ["sample", "sample"],
  [["stratifiedSample", "stratified_sample"], "stratified_sample"],
  [["wordCount", "word_count", "cols.word_count"], "cols.word_count"],
  [["toTimestamp", "to_timestamp", "cols.to_timestamp"], "cols.to_timestamp"],
  [["setColumn", "set", "cols.set"], "cols.set"],
  [["castColumn", "cast", "cols.cast"], "cols.cast"],
  [
    ["setDatatype", "setDtype", "set_dtype", "cols.set_dtype"],
    "cols.set_dtype",
  ],
  [
    ["unsetDatatype", "unsetDtype", "unset_dtype", "cols.unset_dtype"],
    "cols.unset_dtype",
  ],
  ["replace", "cols.replace"],
  [["replaceRegex", "replace_regex"], "cols.replace_regex"],
  ["pattern", "cols.pattern"],
  ["join", "cols.join"],
  ["move", "cols.move"],
  ["sort", "cols.sort"],
  ["item", "cols.item"],
  ["get", "cols.get"],
  ["abs", "cols.abs"],
  ["exp", "cols.exp"],
  ["mod", "cols.mod"],
  ["log", "cols.log"],
  ["ln", "cols.ln"],
  ["pow", "cols.pow"],
  ["sqrt", "cols.sqrt"],
  ["reciprocal", "cols.reciprocal"],
  ["round", "cols.round"],
  ["floor", "cols.floor"],
  ["ceil", "cols.ceil"],
  ["sin", "cols.sin"],
  ["cos", "cols.cos"],
  ["tan", "cols.tan"],
  ["asin", "cols.asin"],
  ["acos", "cols.acos"],
  ["atan", "cols.atan"],
  ["sinh", "cols.sinh"],
  ["cosh", "cols.cosh"],
  ["tanh", "cols.tanh"],
  ["asinh", "cols.asinh"],
  ["acosh", "cols.acosh"],
  ["atanh", "cols.atanh"],
  ["extract", "cols.extract"],
  ["slice", "cols.slice"],
  ["left", "cols.left"],
  ["right", "cols.right"],
  ["mid", "cols.mid"],
  [["toFloat", "to_float"], "cols.to_float"],
  [["toInteger", "to_integer"], "cols.to_integer"],
  [["toBoolean", "to_boolean"], "cols.to_boolean"],
  [["toString", "to_string"], "cols.to_string"],
  ["matchRegex", "cols.match"],
  [["lowercase", "lower"], "cols.lower"],
  [["uppercase", "upper"], "cols.upper"],
  [["propercase", "proper"], "cols.proper"],
  ["title", "cols.title"],
  ["capitalize", "cols.capitalize"],
  ["pad", "cols.pad"],
  ["trim", "cols.trim"],
  [["stripHtml", "strip_html"], "cols.strip_html"],
  [["dateFormat", "date_format"], "cols.date_format"],
  [["wordTokenize", "word_tokenize"], "cols.word_tokenize"],
  [["wordCount", "word_count"], "cols.word_count"],
  ["len", "cols.len"],
  ["reverseString", "cols.reverse"],
  ["remove", "cols.remove"],
  [
    ["normalizeCharacters", "normalizeChars", "normalize_chars"],
    "cols.normalize_chars",
  ],
  [["removeNumbers", "remove_numbers"], "cols.remove_numbers"],
  [["removeWhiteSpaces", "remove_white_spaces"], "cols.remove_white_spaces"],
  [["removeStopWords", "remove_stopwords"], "cols.remove_stopwords"],
  [["removeUrls", "remove_urls"], "cols.remove_urls"],
  [["normalizeSpaces", "normalize_spaces"], "cols.normalize_spaces"],
  [["removeSpecialChars", "remove_special_chars"], "cols.remove_special_chars"],
  [["toDatetime", "to_datetime"], "cols.to_datetime"],
  ["year", "cols.year"],
  ["month", "cols.month"],
  ["day", "cols.day"],
  ["hour", "cols.hour"],
  ["minute", "cols.minute"],
  ["second", "cols.second"],
  ["weekday", "cols.weekday"],
  [["yearsBetween", "years_between"], "cols.years_between"],
  [["lemmatizeVerb", "lemmatize_verb"], "cols.lemmatize_verb"],
  ["impute", "cols.impute"],
  [["fillNa", "fill_na"], "cols.fill_na"],
  ["countCols", "cols.count"],
  ["add", "cols.add"],
  ["sub", "cols.sub"],
  ["mul", "cols.mul"],
  ["div", "cols.div"],
  [["zScore", "z_score"], "cols.z_score"],
  [["modifiedZScore", "modified_z_score"], "cols.modified_z_score"],
  [["minMaxScaler", "min_max_scaler"], "cols.min_max_scaler"],
  [["standardScaler", "standard_scaler"], "cols.standard_scaler"],
  [["maxAbsScaler", "max_abs_scaler"], "cols.max_abs_scaler"],
  ["nest", "cols.nest"],
  ["unnest", "cols.unnest"],
  ["qcut", "cols.qcut"],
  ["cut", "cols.cut"],
  ["clip", "cols.clip"],
  [["stringToIndex", "string_to_index"], "cols.string_to_index"],
  [["indexToString", "index_to_string"], "cols.index_to_string"],
  [["topDomain", "top_domain"], "cols.top_domain"],
  ["domain", "cols.domain"],
  [["URLScheme", "url_scheme"], "cols.url_scheme"],
  [["URLParams", "url_params"], "cols.url_params"],
  [["URLPath", "url_path"], "cols.url_path"],
  ["port", "cols.port"],
  [["URLQuery", "url_query"], "cols.url_query"],
  ["subdomain", "cols.subdomain"],
  [["emailUsername", "email_username"], "cols.email_username"],
  [["emailDomain", "email_domain"], "cols.email_domain"],
  ["missing", "cols.missing"],
  ["fingerprint", "cols.fingerprint"],
  [["nGramFingerprint", "n_gram_fingerprint"], "cols.n_gram_fingerprint"],
  ["metaphone", "cols.metaphone"],
  ["nysiis", "cols.nysiis"],
  [["matchRatingCodex", "match_rating_codex"], "cols.match_rating_codex"],
  [["doubleMethaphone", "double_methaphone"], "cols.double_methaphone"],
  ["soundex", "cols.soundex"],
  ["append", "rows.append"],
  [["greaterThan", "greater_than"], "rows.greater_than"],
  [["greaterThanEqual", "greater_than_equal"], "rows.greater_than_equal"],
  [["lessThan", "less_than"], "rows.less_than"],
  [["lessThanEqual", "less_than_equal"], "rows.less_than_equal"],
  ["equal", "rows.equal"],
  [["notEqual", "not_equal"], "rows.not_equal"],
  ["missing", "rows.missing"],
  ["mismatch", "rows.mismatch"],
  ["match", "rows.match"],
  ["find", "rows.find"],
  ["selectRows", "rows.select"],
  ["sortRows", "rows.sort"],
  ["reverse", "rows.reverse"],
  ["dropRows", "rows.drop"],
  ["between", "rows.between"],
  [["dropNA", "drop_na"], "rows.drop_na"],
  [["dropDuplicates", "drop_duplicates"], "rows.drop_duplicates"],
  ["limit", "rows.limit"],
  [["isIn", "is_in"], "rows.is_in"]
]

alias_map = {}

for alias_group, value in alias_map_list:
    for alias in val_to_list(alias_group):
        alias_map.update({alias: value})

def use_alias(body: dict):

    if body["operation"] in alias_map:
        found = alias_map[body["operation"]]
        if isinstance(found, str):
            body["operation"] = found
        else:
            body.update(found)

    return body


def default_to_engine(body: dict):
    
    if (not "operation" in body) and ("engine" in body):
        body["operation"] = "Optimus"
        
    return body


def default_to_display(body: dict):
    
    if (not "operation" in body) and ("source" in body):
        body["operation"] = "to_dict"
        
    return body


preparers = [
    default_to_engine,
    default_to_display,
    use_alias
]


def prepare(body: dict):
    for preparer in preparers:
        body = preparer(body)

    return body


# from pyspark.sql import functions as F

from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.constants import RELATIVE_ERROR
from optimus.helpers.converter import format_dict
from optimus.helpers.core import one_list_to_val
from optimus.infer import is_numeric
from optimus.outliers.abstract_outliers_threshold import AbstractOutlierThreshold


class ModifiedZScore(AbstractOutlierThreshold):
    """
    Handle outliers from a DataFrame using modified z score
    Reference: http://colingorrie.github.io/outlier-detection.html#modified-z-score-method
    :return:
    """

    def __init__(self, df, col_name, threshold, relative_error=RELATIVE_ERROR):
        """

        :param df:
        :param col_name:
        :param threshold:
        """

        if not is_numeric(threshold):
            raise TypeError("Numeric expected")

        if not is_numeric(relative_error):
            raise TypeError("Numeric expected")

        self.df = df
        self.threshold = threshold
        self.relative_error = relative_error
        self.col_name = one_list_to_val(parse_columns(df, col_name))
        self.z_score = df.cols.modified_z_score(col_name)
        super().__init__()


        # return df.withColumn(m_z_col_name, F.abs(0.6745 * (df[col_name] - mad["median"]) / mad["mad"]))

    def info(self, output: str = "dict"):
        m_z_col_name = name_col(self.col_name, "modified_z_score")

        df = self.df
        z_score = self.z_score

        max_m_z_score = df.rows.select(z_score > self.threshold).cols.max()
        #
        return {"count_outliers": self.count(), "count_non_outliers": self.non_outliers_count(),
                "max_m_z_score": format_dict(max_m_z_score)}



import re

import fastnumbers
import numpy as np
import pandas as pd
from dask_ml.impute import SimpleImputer
from fastnumbers import isintlike, isfloat, isreal, fast_float, fast_int

# From a top point of view we organize Optimus separating the functions in dataframes and dask engines.
# Some functions are commons to pandas and dask.
from optimus.engines.base.ml.contants import STRING_TO_INDEX, INDEX_TO_STRING
from optimus.helpers.columns import parse_columns
from optimus.helpers.constants import Actions
from optimus.helpers.core import val_to_list
from optimus.infer import is_str


def is_string(series):
    def _is_string(value):
        if isinstance(value, str):
            return True
        else:
            return False

    return pd.Series(np.vectorize(_is_string)(series.values).flatten())


def is_integer(series):
    return pd.Series(np.vectorize(isintlike)(series).flatten())


def is_float(series):
    return pd.Series(np.vectorize(isfloat)(series).flatten())


def is_numeric(series):
    return pd.Series(np.vectorize(isreal)(series).flatten())


def word_tokenize(series):
    import nltk
    return nltk.word_tokenize(series)


def is_integer_cudf(series):
    return series.str.isinteger()


def is_float_cudf(series):
    return series.str.isfloat()


def is_numeric_cudf(series):
    return series.str.isnumeric()


def is_string_cudf(series):
    return series.str.isalpha()


# Convertion functions
# cuDF
def to_integer_cudf(series):
    import cudf
    return cudf.to_numeric(series, errors="ignore", downcast="integer")


def to_float_cudf(series):
    import cudf
    return cudf.to_numeric(series, errors="ignore", downcast="float")


def to_string_cudf(series):
    return series.astype(str)


# Vaex

def to_integer_vaex(series):
    return fast_int(series, default=np.nan)


def to_float_vaex(series):
    return fast_float(series, default=np.nan)


def to_string_vaex(series):
    return series.astype(str)


# Pandas Dask
def to_integer(value, *args):
    try:
        # fastnumbers can only handle string or numeric values. Not None, dates or list
        return fastnumbers.fast_forceint(value, default=0)
    except TypeError:
        return np.nan


def to_float(series, *args):
    # if value is None or isinstance(value, str):
    #     return None
    # else:
    return pd.Series(np.vectorize(fast_float)(series, default=np.nan).flatten())

    # try:
    #     # fastnumbers can only handle string or numeric values. Not None, dates or list
    #     return fastnumbers.fast_float(value, default=np.nan)
    # except TypeError:
    #     return np.nan


def to_string(value, *args):
    try:
        return value.astype(str)
    except TypeError:
        return np.nan


def to_boolean(value, *args):
    try:
        # fastnumbers can only handle string or numeric values. Not None, dates or list
        return bool(value)
    except TypeError:
        return np.nan


def to_datetime(value, format):
    return pd.to_datetime(value, format=format, errors="coerce")


def hist(series, bins):
    return np.histogram(series.to_float(), bins=bins)


def to_datetime_cudf(value, format):
    import cudf
    return cudf.to_datetime(value, format=format, errors="coerce")


def impute(df, input_cols, data_type="continuous", strategy="mean", fill_value=None, output_cols=None):
    """

    :param df:
    :param input_cols:
    :param data_type:
    :param strategy:
    # - If "mean", then replace missing values using the mean along
    #   each column. Can only be used with numeric data.
    # - If "median", then replace missing values using the median along
    #   each column. Can only be used with numeric data.
    # - If "most_frequent", then replace missing using the most frequent
    #   value along each column. Can be used with strings or numeric data.
    # - If "constant", then replace missing values with fill_value. Can be
    #   used with strings or numeric data.
    :param output_cols:
    :return:
    """

    imputer = SimpleImputer(strategy=strategy, fill_value=fill_value)

    def _imputer(value):
        return imputer.fit_transform(value.values.reshape(-1, 1))

    if strategy != "most_frequent":
        df = df.cols.to_float(input_cols)

    return df.cols.apply(input_cols, _imputer, output_cols=output_cols, meta_action=Actions.IMPUTE.value,
                         mode="vectorized")


def string_to_index(df, input_cols, output_cols=None, le=None, **kwargs):
    """_
    Maps a string column of labels to an ML column of label indices. If the input column is
    numeric, we cast it to string and index the string values.
    :param df: Dataframe to be transformed
    :param input_cols: Columns to be indexed.
    :param output_cols:Column where the output is going to be saved
    :param le: Label encoder library to make the process
    :return: Dataframe with indexed columns.
    """

    def _string_to_index(value):
        # Label encoder can not handle np.nan
        # value[value.isnull()] = 'NaN'
        return le.fit_transform(value.astype(str))

    return df.cols.apply(input_cols, _string_to_index, output_cols=output_cols,
                         meta_action=Actions.STRING_TO_INDEX.value,
                         mode="vectorized", default=STRING_TO_INDEX)


def index_to_string(df, input_cols, output_cols=None, le=None, **kwargs):
    """
    Maps a column of indices back to a new column of corresponding string values. The index-string mapping is
    either from the ML attributes of the input column, or from user-supplied labels (which take precedence over
    ML attributes).
    :param df: Dataframe to be transformed.
    :param input_cols: Columns to be indexed.
    :param output_cols: Column where the output is going to be saved.
    :param le:
    :return: Dataframe with indexed columns.
    """

    def _index_to_string(value):
        return le.inverse_transform(value)

    return df.cols.apply(input_cols, _index_to_string, output_cols=output_cols,
                         meta_action=Actions.INDEX_TO_STRING.value,
                         mode="vectorized", default=INDEX_TO_STRING)


def find(df, columns, sub, ignore_case=False):
    """
    Find the start and end position for a char or substring
    :param df: Dataframe to be transformed.
    :param columns:
    :param ignore_case:
    :param sub:
    :return:
    """

    columns = parse_columns(df, columns)
    sub = val_to_list(sub)

    def get_match_positions(_value, _separator):
        result = None
        if is_str(_value):
            # Using re.IGNORECASE in finditer not seems to work
            if ignore_case is True:
                _separator = _separator + [s.lower() for s in _separator]
            regex = re.compile('|'.join(_separator))

            length = [[match.start(), match.end()] for match in
                      regex.finditer(_value)]
            result = length if len(length) > 0 else None
        return result

    dfd = df.data

    for col_name in columns:
        # Categorical columns can not handle a list inside a list as return for example [[1,2],[6,7]].
        # That could happened if we try to split a categorical column
        # dfd[col_name] = dfd[col_name].astype("object")
        dfd[col_name + "__match_positions__"] = dfd[col_name].astype("object").apply(get_match_positions,
                                                                                     args=(sub,))

    return df.new(dfd)
